
/**********************************************************************************
*  File : linux/arch/arm/kernel/gdb-stub.c
*
*  This file contains the remote protocol implementation on the target side.
*
***********************************************************************************/


/***************** The gdb remote communication protocol****************************
*
*   A debug packet whose contents are <data>
*   is encapsulated for transmission in the form:
*
*        $ <data> # CSUM1 CSUM2
*
*        <data> must be ASCII alphanumeric and cannot include characters
*        '$' or '#'.  If <data> starts with two characters followed by
*        ':', then the existing stubs interpret this as a sequence number.
*
*        CSUM1 and CSUM2 are ascii hex representation of an 8-bit
*        checksum of <data>, the most significant nibble is sent first.
*        the hex digits 0-9,a-f are used.
*
*   Receiver responds with:
*
*        +       - if CSUM is correct and ready for next packet
*        -       - if CSUM is incorrect
*
*   <data> is as follows:
*   All values are encoded in ascii hex digits.
*
*        Request         Packet
*
*        read registers  g
*        reply           XX....X         Each byte of register data
*                                        is described by two hex digits.
*                                        Registers are in the internal order
*                                        for GDB, and the bytes in a register
*                                        are in the same order the machine uses.
*                        or ENN          for an error.
*
*       write regs      GXX..XX         Each byte of register data
*                                        is described by two hex digits.
*        reply           OK              for success
*                        ENN             for an error
*
*        write reg       Pn...=r...      Write register n... with value r...,
*                                        which contains two hex digits for each
*                                        byte in the register (target byte
*                                        order).
*        reply           OK              for success
*                        ENN             for an error
*        (not supported by all stubs).
*
*        read mem        mAA..AA,LLLL    AA..AA is address, LLLL is length.
*        reply           XX..XX          XX..XX is mem contents
*                                        Can be fewer bytes than requested
*                                        if able to read only part of the data.
*                        or ENN          NN is errno
*
*        write mem       MAA..AA,LLLL:XX..XX
*                                        AA..AA is address,
*                                        LLLL is number of bytes,
*                                        XX..XX is data
*        reply           OK              for success
*                        ENN             for an error (this includes the case
*                                        where only part of the data was
*                                        written).
*
*        write mem       XAA..AA,LLLL:XX..XX
*         (binary)                       AA..AA is address,
*                                        LLLL is number of bytes,
*                                        XX..XX is binary data
*        reply           OK              for success
*                        ENN             for an error
*
*        cont            cAA..AA         AA..AA is address to resume
*                                        If AA..AA is omitted,
*                                        resume at same address.
*
*        step            sAA..AA         AA..AA is address to resume
*                                        If AA..AA is omitted,
*                                        resume at same address.
*
*        last signal     ?               Reply the current reason for stopping.
*                                        This is the same reply as is generated
*                                        for step or cont : SAA where AA is the
*                                        signal number.
*
*        There is no immediate reply to step or cont.
*        The reply comes when the machine stops.
*        It is           SAA             AA is the "signal number"
*
*        or...           TAAn...:r...;n:r...;n...:r...;
*                                        AA = signal number
*                                        n... = register number
*                                        r... = register contents
*       or...           WAA             The process exited, and AA is
*                                       the exit status.  This is only
*                                       applicable for certains sorts of
*                                        targets.
*        kill request    k
*
*       toggle debug    d               toggle debug flag.
*        reserved        <other>         On other requests, the stub should
*                                        ignore the request and send an empty
*                                        response ($#<checksum>).  This way
*                                        we can extend the protocol and GDB
*                                        can tell whether the stub it is
*                                        talking to uses the old or the new.
*        search          tAA:PP,MM       Search backwards starting at address
*                                        AA for a match with pattern PP and
*                                        mask MM.  PP and MM are 4 bytes.
*                                        Not supported by all stubs.
*
*        general query   qXXXX           Request info about XXXX.
*        general set     QXXXX=yyyy      Set value of XXXX to yyyy.
*        query sect offs qOffsets        Get section offsets.  Reply is
*                                        Text=xxx;Data=yyy;Bss=zzz
*        console output  Otext           Send text to stdout.  Only comes from
*                                        remote target.
*
*        Responses can be run-length encoded to save space.  A '*' means that
*        the next character is an ASCII encoding giving a repeat count which
*        stands for that many repititions of the character preceding the '*'.
*        The encoding is n+29, yielding a printable character where n >=3
*        (which is where rle starts to win).  Don't use an n > 126.
*
*        So
*        "0* " means the same as "0000".
*
**********************************************************************************/

#include "gdb-main.h"

#if !defined(GDB_RXBUFLEN)
#define GDB_RXBUFLEN 200
#endif

#define min_val(a,b) ((a) > (b) ? (b) : (a))
#define max_val(a,b) ((a) > (b) ? (a) : (b))

#define is_aligned_long(addr,len) \
  (((len) >= sizeof(long)) && ((long)(addr) % sizeof(long) == 0))

#define is_aligned_short(addr,len) \
  (((len) >= sizeof(short)) && ((long)(addr) % sizeof(short) == 0))

/* Thread support added */
#ifdef CONFIG_GDB_THREAD
	#include <linux/spinlock.h>

	#define MAX_PACKET_LENGTH (400 -1)
	#define BUF_THREAD_ID_SIZE 16
	#define BUFMAX 2048
	typedef unsigned char threadref[8];
	static int get_threadid_from_pkt(char**, int*);
	static struct task_struct * get_thread(int);
	static void int_to_threadref(threadref*, const int);
	static char * pack_threadid(char*, threadref*);
	static char * pack_hex_byte(char*,int);
	static void set_thread_msg(char*);
	static void thread_status_msg(char*);
	static void query_msg(char*);
	static void thread_id_msg(void);
	static void thread_info_msg(void);
	static void thread_extra_info_msg(char*);

	static struct task_struct *trapped_thread;
	struct task_struct *current_thread;

	static int nThreadCount =0; /* keep track of number of threads traversed in case of 
								   'qsThreadInfo' and 'qsThreadInfo' packets */
	static int qsThreadInfo =0; /* to identify if the packet is 'qsThreadInfo' */

#endif

/******************************************************************************
*  FUNCTION:    hex_to_int                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Converts '[0-9,a-f,A-F]' to its integer equivalent                                         
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hex character whose interger equivalent is to be found                                                                   
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     integer equivalent    
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static int
hex_to_int (char h)
{
	if (h >= 'a' && h <= 'f')
		return h - 'a' + 10;
	if (h >= '0' && h <= '9')
		return h - '0';
	if (h >= 'A' && h <= 'F')
		return h - 'A' + 10;
	
	return -1;
}


/******************************************************************************
*  FUNCTION:    lnibble_to_hex                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Converts the low nibble of input to its hex character equivalent                                         
*                                                                            
*  INPUT PARAMETERS:                                                         
*     nibble whose hex equivalent is to be found                                                                   
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     hex equivalent    
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static char
lnibble_to_hex (char i)
{
	static const char lnibble_to_hex_table[] = "0123456789abcdef";
	return lnibble_to_hex_table[i & 0xf];
}


/******************************************************************************
*  FUNCTION:    hargs_parse_long                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Translates a delimited hex string to a long                                         
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hargs : string from the packet
*     l     : long value 
*     delim : delimiter which separates the parts of packet                                                                   
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     Pointer to next character of the packet after processing the hex string    
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static const char *
hargs_parse_long (const char *hargs, long *l, int delim)
{
	*l = 0;
	while (*hargs != delim)
		*l = (*l << 4) + hex_to_int (*hargs++);
	
	return hargs + 1;
}


/******************************************************************************
*  FUNCTION:    mem_to_hexbuf                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Converts a memory region of length len bytes, starting at mem, into
*     a string of hex bytes.                                        
*                                                                            
*  INPUT PARAMETERS:                                                         
*     mem  : memory to be read
*     hbuf : memory contents will be placed in this
*     len  : number of bytes to be read 
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     Returns the number of bytes placed into hexbuf    
*                                                                            
*  NOTES:                                                                    
*     the lcbuf unions assume and depend that lbuf and sbuf start at
*     the same address.  Is this always correct?  Is there a better way?                                                                 
*
*     This function carefully preserves the endianness of the data,
*     because that's what gdb expects.  This function also optimizes the
*     read process into the largest units possible, in case we're reading
*     a peripheral register that can't deal with unaligned or byte-wide
*     accesses.
******************************************************************************/
static int
mem_to_hexbuf (const void *mem, char *hbuf, int len)
{
	int i = 0;
	union
	{
		long lbuf;
		short sbuf;
		char cbuf[sizeof (long)];
	} lcbuf;
	
	int cbuflen;
	int retval = 0;
	long * mem_l = (long*)mem;
	short *mem_s = (short*)mem;
	char *mem_c = (char*)mem;


	while (len > 0)
    {
		if (is_aligned_long (mem, (unsigned)len))
		{
			cbuflen = sizeof (long);
			lcbuf.lbuf = *mem_l;
			mem_l++;
			len -= sizeof (long);
		}
		else if (is_aligned_short (mem, (unsigned)len))
		{
			cbuflen = sizeof (short);
			lcbuf.sbuf = *mem_s;
			mem_s++;
			len -= sizeof (short);
		}
		else
		{
			cbuflen = sizeof (char);
			lcbuf.cbuf[0] = *mem_c;
			mem_c++;
			len -= sizeof (char);
		}
		
		for (i = 0; i < cbuflen; i++)
		{
			*hbuf++ = lnibble_to_hex (lcbuf.cbuf[i] >> 4);
			*hbuf++ = lnibble_to_hex (lcbuf.cbuf[i]);
			retval += 2;
		}
    }

	return retval;
}


/******************************************************************************
*  FUNCTION:    hexbuf_to_mem                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Reads (len * 2) hex digits from hbuf, converts them to binary,
*     writes them to mem. Returns a pointer to the first empty byte after
*     the region written.                                         
*                                                                            
*  INPUT PARAMETERS:                                                         
*     mem  : memory to be written
*     hbuf : memory whose contents are to be placed in mem
*     len  : number of bytes to be written                                                                   
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     pointer to the first empty byte after the region written    
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static char *
hexbuf_to_mem (const char *hbuf, void *mem, int len)
{
	unsigned int i = 0;
	long * mem_l = (long*)mem;
	short *mem_s = (short*)mem;
	char *mem_c = (char*)mem;
	
	union
	{
		long lbuf;
		short sbuf;
		char cbuf[sizeof (long)];
	} lcbuf;
	
	/*  void *cache_start = mem;
		int cache_len = len;  */

	while (len > 0)
    {
		if (is_aligned_long (mem, (unsigned)len))
		{
			for (i = 0; i < sizeof (long); i++)
			{
				lcbuf.cbuf[i] = (hex_to_int (*hbuf++) << 4);
				lcbuf.cbuf[i] += hex_to_int (*hbuf++);
			}
			
			*mem_l++ = lcbuf.lbuf;
			len -= sizeof (long);
			mem = mem_l;
		}
		else if (is_aligned_short (mem, (unsigned)len))
		{
			for (i = 0; i < sizeof (short); i++)
			{
				lcbuf.cbuf[i] = (hex_to_int (*hbuf++) << 4);
				lcbuf.cbuf[i] += hex_to_int (*hbuf++);
			}
			*mem_s++ = lcbuf.sbuf;
			len -= sizeof (short);
			mem = mem_s;
		}
		else
		{
			lcbuf.cbuf[0] = (hex_to_int (*hbuf++) << 4);
			lcbuf.cbuf[0] += hex_to_int (*hbuf++);
			*mem_c++ = lcbuf.cbuf[0];
			len -= sizeof (char);
			mem = mem_c;
		}
    }
	
	return mem;
}

/******************************************************************************
*  FUNCTION:    xbin_to_bin                                  
*                                                                            
*  DESCRIPTION:                                                              
*                                              
*                                                                            
*  INPUT PARAMETERS:                                                         
*
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static const void *
xbin_to_bin (const void *xbin, char *bin)
{
	char* xbin_c = (char*)xbin;
	if (*xbin_c == 0x7d)
    {
		xbin_c++;
		*bin = *(xbin_c) ^ 0x20;
		xbin_c++;
    }
	else
		*bin = *xbin_c++;

	return xbin;
}


/******************************************************************************
*  FUNCTION:    xmem_to_mem                                  
*                                                                            
*  DESCRIPTION:                                                              
*   Converts the escaped-binary ('X' packet) array pointed to by buf
*   into binary, to be placed in mem. Returns a pointer to the first
*   empty byte after the region written.                                           
*                                                                            
*  INPUT PARAMETERS:                                                         
*   xmem  : data to be written
*   mem   : address at which data is to be written
*   len   :  length of data
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*   None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*	pointer to the first empty byte after the region written.  
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static char *
xmem_to_mem (const char *xmem, void *mem, int len)
{
	unsigned int i = 0;
	long *mem_l = (long*)mem;
	short *mem_s = (short*)mem;
	char *mem_c = (char*)mem;
	union
	{
		long lbuf;
		short sbuf;
		char cbuf[sizeof (long)];
	} lcbuf;

	while (len > 0)
	{
		if (is_aligned_long (mem, (unsigned)len))
		{
			for (i = 0; i < sizeof (long); i++)
				xmem = xbin_to_bin (xmem, &lcbuf.cbuf[i]);

			*mem_l++ = lcbuf.lbuf;
			len -= sizeof (long);
			mem = mem_l;
		}
		else if (is_aligned_short (mem, (unsigned)len))
		{
			for (i = 0; i < sizeof (short); i++)
				xmem = xbin_to_bin (xmem, &lcbuf.cbuf[i]);

			*mem_s++ = lcbuf.sbuf;
			len -= sizeof (short);
			mem = mem_s;
		}
		else
		{
			xmem = xbin_to_bin (xmem, &lcbuf.cbuf[0]);
			*mem_c++ = lcbuf.cbuf[0];
			len -= sizeof (char);
			mem = mem_c;
		}
	}
	
	return mem;
}


/******************************************************************************
*  FUNCTION:    putbuf                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Writes a buffer of length len to gdb_putc().
*                                                                            
*  INPUT PARAMETERS:                                                         
*     len : number of bytes to be written
*     buf : buffer to be written to uart
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     Returns the checksum of the bytes.                                           
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static int
putbuf (int len, const char *buf)
{
	unsigned char sum = 0;

	while (len--)
    {
		sum += *(unsigned char *) buf;
		gdb_putc (*buf++);
    }
	
	return sum;
}


/******************************************************************************
*  FUNCTION:    putmsg                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Prepares a packet and writes it to UART.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     c     : first character to be sent after '$'. can be empty
*     buf   : buffer containing data to be written to uart
*     len   : number of bytes to be written 
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static void putmsg (char c, const char *buf, int len)
{
	unsigned char sum;
	int wait = 0;

	do
	{
		/* send the header */
		gdb_putc('$');
		
		/* send the message type, if specified */
		if (c) gdb_putc(c);

		/* send the data */
		sum = c + putbuf(len, buf);

		/* send the footer */
		gdb_putc('#');
		gdb_putc(lnibble_to_hex(sum >> 4));
		gdb_putc(lnibble_to_hex(sum));
  
		wait = 1000;

		while(wait>0) wait--;
 
		} while ('+' != gdb_getc());
	
	return;
}


/******************************************************************************
*  FUNCTION:    getmsg                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Get the packet from UART.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     rxbuf   : buffer into which the data is received	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     status as 1                                        
*                                                                            
*  NOTES:                                                                    
*     None                                                                   
*                                                                            
******************************************************************************/
static int
getmsg (char *rxbuf)
{
	char c;
	unsigned char sum;
	unsigned char rx_sum;
	char *buf;

get_msg:
	
	/* wait around for start character, ignore all others */
	while (gdb_getc () != '$');

	/* start counting bytes */
	buf = rxbuf;
	sum = 0;

	/* read until we see the '#' at the end of the packet */
	do
    {
		*buf++ = c = gdb_getc ();
		if (c != '#')
		sum += c;

		/* since the buffer is ascii, may as well terminate it */
		*buf = 0;
    }while (c != '#');
	
	/* receive checksum */
	rx_sum = hex_to_int (gdb_getc ());
	rx_sum = (rx_sum << 4) + hex_to_int (gdb_getc ());

	/* if computed checksum doesn't match received checksum, then reject */
	if (sum != rx_sum)
    {
		gdb_putc ('-');
		goto get_msg;
    }

	/* got the message ok */
	else
		gdb_putc ('+');

  return 1;
}


/******************************************************************************
*  FUNCTION:    last_signal                                  
*                                                                            
*  DESCRIPTION:                                                              
*     It preapres a packet and sent it to the host to tell the reason why target
*     halted.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     rxbuf   : buffer into which the data is received	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void
last_signal (int sigval)
{
#ifndef CONFIG_GDB_THREAD
	char tx_buf[2];
  
	tx_buf[0] = lnibble_to_hex (sigval >> 4);
	tx_buf[1] = lnibble_to_hex (sigval);
	putmsg ('S', tx_buf, 2);
#else
	int threadid;
	threadref thref;
	//char *out = remcomOutBuffer;
	char tx_buf[BUFMAX];
	char *p = &tx_buf[0];
	const char *tstring = "thread";
	int len =0;

	//*out++ = 'T';
	*p++ = lnibble_to_hex (sigval >> 4);
	*p++ = lnibble_to_hex (sigval);

	while (*tstring) {
                *p++ = *tstring++;
	}
	*p++ = ':';

	threadid = trapped_thread->pid;
	if (threadid == 0) threadid = PID_MAX_DEFAULT;
	int_to_threadref(&thref, threadid);
	pack_threadid(p, &thref);
	p += BUF_THREAD_ID_SIZE;
	*p++ = ';';

//	*p = 0;

	len = (int)p - (int)&tx_buf[0];
	putmsg('T', &tx_buf[0], len);
#endif
}


#if 0

/*
  "expedited response" message
  "Txx..........."
*/
static void
expedited (int sigval)
{
	long val;
	int id = 0;
	int reglen;
	int sum __unused;
	char tx_buf[256] = {0};
	int len;  
	char *p = &tx_buf[0];

	do
	{
		*p++ = lnibble_to_hex (sigval >> 4); 
		*p++ = lnibble_to_hex (sigval);

		/* register values */
		id = 0;
		while ((reglen = gdb_peek_register_file (id, &val)) != 0)
		{
			if(id==0x10) break;
			
			/* get the ID.  */
			*p++ = lnibble_to_hex (id >> 4);
			*p++ = lnibble_to_hex (id);
			*p++ = ':';
			
			/* register value */
			switch (reglen)
            {
            case 4:
              *p++ = lnibble_to_hex (val >> 28);
              *p++ = lnibble_to_hex (val >> 24);
            case 3:
              *p++ = lnibble_to_hex (val >> 20);
              *p++ = lnibble_to_hex (val >> 16);
            case 2:
              *p++ = lnibble_to_hex (val >> 12);
              *p++ = lnibble_to_hex (val >> 8);
            case 1:
              *p++ = lnibble_to_hex (val >> 4);
              *p++ = lnibble_to_hex (val);
              break;
            }
			
			*p++ = ';';
			id++;
		}

    }while(0);

    len = (int)p - (int)&tx_buf[0];
    putmsg('T', &tx_buf[0], len);
	
	while ('+' != gdb_getc ());
	
	return;
}
#endif 


/******************************************************************************
*  FUNCTION:    read_memory                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Handles the memory read packet. ('m')
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hargs   : packet from the host.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:
*  Address can be invalid address(memory fault), in that case fault handler 
*  is invoked which prints message and ignores it. 
*                                                                            
******************************************************************************/
static void
read_memory (const char *hargs)
{
	char tx_buf[sizeof (long) * 2];
	long addr = 0, orig_addr = 0;
	long len = 0, orig_len = 0;
	int tx;
	unsigned char sum = 0;
	
	/* parse address, length */
	hargs = hargs_parse_long (hargs, &addr, ',');
	hargs = hargs_parse_long (hargs, &len, '#');

	orig_addr = addr;
	orig_len = len;

	if (gdb_setjmp(fault_jmp_buf) == 0)
	{
		gdb_fault_expected = 1;
		
		do
		{
			addr = orig_addr;
			len = orig_len;

			gdb_putc ('$');
		  
			/* send the message a piece at a time, so we don't need much memory */
			while (len)
			{
				tx =  mem_to_hexbuf ((void *) addr, tx_buf, min_val (len, (int)sizeof (long)));
				sum += putbuf (tx, tx_buf);
				addr += tx / 2;
				len -= min_val (tx / 2, len);
			}
			
			gdb_putc ('#');
			gdb_putc (lnibble_to_hex (sum >> 4));
			gdb_putc (lnibble_to_hex (sum));
		}while (gdb_getc() != '+');
	}
	else
	{
		putmsg(0, "E01", 3);
	}

	gdb_fault_expected = 0;
}

/******************************************************************************
*  FUNCTION:    write_memory                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Handles the write memory packet. ('M')
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hargs   : packet from the host.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*  Address can be invalid address(memory fault), in that case fault handler 
*  is invoked which prints message and ignores it. 
*                                                                            
******************************************************************************/
static void
write_memory (const char *hargs)
{
	long addr = 0;
	long len = 0;
	
	/* parse address, length */
	hargs = hargs_parse_long (hargs, &addr, ',');
	hargs = hargs_parse_long (hargs, &len, ':');
	
	if (gdb_setjmp(fault_jmp_buf) == 0)
	{
	   gdb_fault_expected = 1;
	   
	   /* write all requested bytes */
	   hexbuf_to_mem (hargs, (void *) addr, len);
	   putmsg (0, "OK", 2);
	}
	else
	{
		putmsg(0, "E01", 3);
	}

	gdb_fault_expected = 0;
}

/******************************************************************************
*  FUNCTION:    write_xbin_memory                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Handles the write memory packet when the data recieved is in binary.('X')
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hargs   : packet from the host.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void
write_xbin_memory (const char *hargs)
{
	long addr = 0;
	long len = 0;
	
	/* parse address, length */
	hargs = hargs_parse_long (hargs, &addr, ',');
	hargs = hargs_parse_long (hargs, &len, ':');
	
	/* write all requested bytes */
	xmem_to_mem (hargs, (void *) addr, len);
	
	putmsg (0, "OK", 2);
	return;
}


/******************************************************************************
*  FUNCTION:    write_registers                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Handles the write registers packet ('G');
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hargs   : packet from the host.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void
write_registers (char *hargs)
{
	int id = 0;
	long val;
	int reglen;
	
	while (*hargs != '#')
    {
		/* how big is this register? */
		reglen = gdb_peek_register_file (id, &val);
		
		if (reglen)
		{
			/* extract the register's value */
			hexbuf_to_mem (hargs, &val, reglen);
			hargs += sizeof (long) * 2;
			
			/* stuff it into the register file */
			gdb_poke_register_file (id++, val);
		}
		
		else
			break;
    }
	
	putmsg (0, "OK", 2);
	return;
}


/******************************************************************************
*  FUNCTION:    read_registers                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Handles the read registers packet ('g');
*                                                                            
*  INPUT PARAMETERS:                                                         
*
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void
read_registers (void)
{
	char tx_buf[sizeof (long) * 2];
	long val;
	int id = 0;
	int reglen;
	unsigned char sum;
	
	do
    {
		gdb_putc ('$');
		sum = 0;
		
		/* send register values */
		id = 0;
		while ((reglen = gdb_peek_register_file (id++, &val)) != 0)
			sum += putbuf (mem_to_hexbuf (&val, tx_buf, reglen), tx_buf);
		
		/* send the message footer */
		gdb_putc ('#');
		gdb_putc (lnibble_to_hex (sum >> 4));
		gdb_putc (lnibble_to_hex (sum));
	}
	while (0);

	while ('+' != gdb_getc ());
	
	return;
}

/******************************************************************************
*  FUNCTION:    write_register                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Handles the write register packet ('P');
*                                                                            
*  INPUT PARAMETERS:                                                         
*     hargs   : packet from the host.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void
write_register (char *hargs)
{
	long id = 0;
	long val = 0;
	int reglen;
	
	while (*hargs != '=')
		id = (id << 4) + hex_to_int (*hargs++);
	
	hargs++;
	
	reglen = gdb_peek_register_file (id, &val);
	hexbuf_to_mem (hargs, &val, reglen);
	gdb_poke_register_file (id, val);
	putmsg (0, "OK", 2);

	return;
}

#if 0
void
gdb_console_output (int len, const char *buf)
{
	char tx_buf[2];
	unsigned char sum;
	
	gdb_putc ('$');
	sum = putbuf (1, "O");
	
	while (len--)
    {
		tx_buf[0] = lnibble_to_hex (*buf >> 4);
		tx_buf[1] = lnibble_to_hex (*buf++);
		sum += putbuf (2, tx_buf);
    }
	
	/* send the message footer */
	gdb_putc ('#');
	gdb_putc (lnibble_to_hex (sum >> 4));
	gdb_putc (lnibble_to_hex (sum));
	
	/* DON'T wait for response; we don't want to get hung
     up here and halt the application if gdb has gone away! */
	return;
}
#endif


/******************************************************************************
*  FUNCTION:    gdb_monitor                                  
*                                                                            
*  DESCRIPTION:                                                              
*     This is the gdb command processor. All command processing funtions are 
*     called from this function.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     sigval   : GDB signal number.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     gdb_connected : GDB status is now connected                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
void
gdb_monitor (int sigval)
{
	char rxbuf[GDB_RXBUFLEN];
	char *hargs;
	long addr;
  
	/* for Zpacket support.  */ 
	long len;
	long type;
	int err=0;
	int is_Z = 0;

	gdb_monitor_onentry ();

	while (1)
    {
		getmsg (rxbuf);
		
		if(!gdb_connected)
		{
			printk("GDB connected, Remote Debugging with GDB...\n");
			gdb_connected = 1;
		}

		hargs = rxbuf;
		switch (*hargs++)
		{
		case '?':
			last_signal (sigval);
			break;

		case 'C':
			/* continue with signal, we ignore the signal num.  */
			gdb_continue(0);
			break;

		case 'c':
			/* this call probably doesn't return */
			hargs_parse_long (hargs, &addr, '#');
			gdb_continue (addr);
			/* if it does, exit back to interrupted code */
			return;

		case 'd':
			/* toggle debug flag */
			gdb_connected = !(gdb_connected);
			break;

		case 'D':
			/* detach from target, gdb is going away */
			putmsg (0, "OK", 2);
			gdb_detach ();
			break;

		case 'g':
			read_registers ();
			break;

		case 'G':
			#ifdef CONFIG_GDB_THREAD
			if (current_thread) 
			{
				/* Set register contents - currently can't set other thread's registers */
				putmsg(0, "E01", 3);
				break;	
			}			
			#endif

			write_registers (hargs);
			break;

		case 'H':
			/* set thread--- unimplemented, but gdb likes it */
			#ifdef CONFIG_GDB_THREAD
				set_thread_msg(hargs); // added for thread support
			#else	
				putmsg (0, "OK", 2);
			#endif
			break;

		case 'k':
			/* kill program */
			gdb_connected = 0;
			gdb_kill ();
			break;

		case 'm':
			read_memory (hargs);
			break;

		case 'M':
			write_memory (hargs);
			break;

		case 'P':
			#ifdef CONFIG_GDB_THREAD
			if (current_thread) 
			{
				/* Set register contents - currently can't set other thread's registers */
				putmsg(0, "E01", 3);
				break;	
			}			
			#endif

			write_register (hargs);
			break;

		case 'S':
			/* step with signal. we ignore the signal num.  */
			gdb_step(0);
			break;	

		case 's':
			/* step (address optional) */
			hargs_parse_long (hargs, &addr, '#');
			gdb_step (addr);
			/* exit back to interrupted code */
			return;

#ifdef CONFIG_GDB_THREAD
		case 'T':	/* Query thread status */
			thread_status_msg(hargs);
			break;

		case 'q':	/* Handle query - currently thread-related */
			query_msg(hargs);
			break;
#endif

		case 'X':
			/* write to memory (source in escaped-binary format) */
			write_xbin_memory (hargs);
			break;

		case 'Z':
			is_Z = 1;
		case 'z':	   
			addr = 0;
			/* insert break or watch point. */
			type = hex_to_int(*hargs++);
			if(*hargs++!=',') 
			{
				/* Error in packet */
				putmsg(0, "E02", 3);
				break;
			}
			
			hargs_parse_long(hargs, &addr, ',');
			hargs_parse_long(hargs, &len, '#');
          
			if(!addr)
			{
				/* memory fault */
				putmsg(0, "E02", 3);
				break;
			}
			switch (type)
			{
			/* software breakpoint.  */
			case 0:
				if(is_Z)
				{
					err = insert_breakpoint(addr, len);
					/* memory fault */
					if(err!=0)
					{
						putmsg(0, "E01", 3);
						break;
					}
				}
				else
				{
					/* memory fault */
					err =  remove_breakpoint(addr, len);
					if(err!=0)
					{
						putmsg(0, "E01", 3);
						break;
					}
				}
				
				putmsg (0, "OK", 2);
				break;
			
			default:
				putmsg (0, "", 0);	
			}
			
			break;
			
		default:
		/* received a command we don't recognize---
			send empty response per gdb spec */
			putmsg (0, "", 0);	  
		}
	}
	
	return;
}


/******************************************************************************
*  FUNCTION:    gdb_handle_exception                                  
*                                                                            
*  DESCRIPTION:                                                              
*     This function is called from breakpoint_handler function to do furthur processing.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     sigval   : GDB signal number.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
void
gdb_handle_exception (int sigval)
{
/* Set the current thread which will execute */
#ifdef CONFIG_GDB_THREAD
   /* Until GDB specifies a thread */
   current_thread = NULL;
   trapped_thread = current;
#endif

  /* Tell the host why the target halted i.e signal, current thread id etc.*/
  if(gdb_connected)
  {
	  last_signal (sigval);  
  }

  /* ask gdb what to do next */
  gdb_monitor (sigval);

  /* return to the interrupted code */
  //gdb_return_from_exception ();
}


/********* THREAD SUPPORT ******/

#ifdef CONFIG_GDB_THREAD

/******************************************************************************
*  FUNCTION:    set_thread_msg                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Set the status for a thread 
*                                                                            
*  INPUT PARAMETERS:                                                         
*     sigval   : GDB signal number.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     current_thread : current thread id is changed according to 'Hg' packet.                                                       
*                                                                            
*  RETURN VALUES:                                                            
*     None                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void set_thread_msg(char *pkt)
{
#ifndef CONFIG_GDB_THREAD
	putmsg (0, "OK", 2);
#else
	int threadid;
	struct task_struct *thread = NULL;
	char *ptr;

	switch (*pkt++) {

       	/* To select which thread for gG etc messages, i.e. supported */
	case 'g':
		ptr = pkt;
		get_threadid_from_pkt(&ptr, &threadid);
		thread = get_thread(threadid);

		/* If we haven't found it */
		if (!thread) {
			putmsg(0, "E01", 3);
			break;
		}

		/* Set current_thread (or not) */
		if (thread == trapped_thread)
		{
			current_thread = NULL;
		}
		else
		{
			current_thread = thread;			
		}
		putmsg (0, "OK", 2);
		break;

	/* To select which thread for cCsS messages, i.e. unsupported */
	case 'c':
		putmsg (0, "OK", 2);
		break;

	default:
		putmsg (0, "", 0);
		break;
	}
#endif
}


/******************************************************************************
*  FUNCTION:    get_threadid_from_pkt                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Extract the thread id from the packet. 
*                                                                            
*  INPUT PARAMETERS:                                                         
*     ptr : packet from host.	
*     intvalue : value to be returned
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     number of characters processed                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static int get_threadid_from_pkt(char **ptr, int *intValue)
{
	int numChars = 0;
	int hexValue;

	*intValue = 0;

	while (**ptr) {
		hexValue = hex_to_int(**ptr);
		if (hexValue < 0)
			break;

		*intValue = (*intValue << 4) | hexValue;
		numChars ++;

		(*ptr)++;
	}

	return (numChars);
}

/******************************************************************************
*  FUNCTION:    get_thread                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Get the task structure for a thread. 
*                                                                            
*  INPUT PARAMETERS:                                                         
*     pid : Thread id of the thread whose structure is required.	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     task structure for a particular pid                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static struct task_struct * get_thread(int pid)
{
	struct task_struct *thread;

	/* Use PID_MAX_DEFAULT w/gdb for pid 0 */
	if (pid == PID_MAX_DEFAULT) pid = 0;

	/* First check via PID */
	thread = find_task_by_pid(pid);

	if (thread)
		return thread;

	/* Start at the start */
	//thread = init_tasks[0];
	thread = (&init_task);

	/* Walk along the linked list of tasks */
	do {
		if (thread->pid == pid)
			return thread;
		//thread = thread->next_task;
		thread = list_entry((thread)->tasks.next, struct task_struct, tasks);
	} while (thread != (&init_task));
	
	return NULL;
}

/******************************************************************************
*  FUNCTION:    int_to_threadref                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Get the thread id into 'threadref' form(so that it can be stored in the packet). 
*                                                                            
*  INPUT PARAMETERS:                                                         
*     id : thread id will be put into this buffer.
*     value : thread number	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     task structure for a particular pid                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void int_to_threadref(threadref * id, const int value)
{
	unsigned char *scan = (unsigned char *) id;
	int i = 4;

	while (i--)
		*scan++ = 0;

	*scan++ = (value >> 24) & 0xff;
	*scan++ = (value >> 16) & 0xff;
	*scan++ = (value >> 8) & 0xff;
	*scan++ = (value & 0xff);
}


/******************************************************************************
*  FUNCTION:    pack_threadid                                  
*                                                                            
*  DESCRIPTION:                                                              
*    store the thread id into the packet. 
*                                                                            
*  INPUT PARAMETERS:                                                         
*     pkt : packet into which the thred id will be put.
*     id  : thread id 	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     packet with thread id                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static char * pack_threadid(char *pkt, threadref * id)
{
	char *limit;
	unsigned char *altid;

	altid = (unsigned char *) id;

	limit = pkt + BUF_THREAD_ID_SIZE;
	while (pkt < limit)
		pkt = pack_hex_byte(pkt, *altid++);
	return pkt;
}


/******************************************************************************
*  FUNCTION:    pack_hex_byte                                  
*                                                                            
*  DESCRIPTION:                                                              
*    utility function to store threadid into packet. called from 'pack_threadid'.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     pkt : packet into which the thred id will be put.
*     byte  : individual byte of the thread id 	
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     packet with thread id(input byte)                                        
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static char * pack_hex_byte(char *pkt, int byte)
{
	*pkt++ = lnibble_to_hex((byte >> 4) & 0xf);
	*pkt++ = lnibble_to_hex((byte & 0xf));
	return pkt;
}


/******************************************************************************
*  FUNCTION:    thread_status_msg                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Handles query packet to check if the thread is alive.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     pkt : query packet from host.
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*     None   
*                                                                    
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void thread_status_msg(char *pkt)
{
	char *ptr;
	int threadid;
	struct task_struct *thread = NULL;

	ptr = pkt;
	get_threadid_from_pkt(&ptr, &threadid);
	thread = get_thread(threadid);

	if (thread)
		putmsg (0, "OK", 2);
	else
		putmsg(0, "E01", 3);
}

/******************************************************************************
*  FUNCTION:    thread_id_msg                                  
*                                                                            
*  DESCRIPTION:                                                              
*     Send the current thread ID.('T' packet)
*                                                                            
*  INPUT PARAMETERS:                                                         
*     None
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:  
*     None                                                          
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
/* Send the current thread ID */
static void thread_id_msg(void)
{
	int threadid;
	threadref thref;

	char tx_buf[BUFMAX];
	char *p = &tx_buf[0];
	int len;


	*p++ = 'Q';
	*p++ = 'C';

	if (current_thread)
		threadid = current_thread->pid;
	else if (trapped_thread)
		threadid = trapped_thread->pid;
	else /* Impossible, but just in case! */
	{
		putmsg(0, "E01", 3);
		return;
	}

	/* Translate pid 0 to PID_MAX_DEFAULT for gdb */
	if (threadid == 0) threadid = PID_MAX_DEFAULT;

	int_to_threadref(&thref, threadid);
	pack_threadid(p, &thref);
	p[BUF_THREAD_ID_SIZE] = '\0';
	p += BUF_THREAD_ID_SIZE;

	len = (int)p - (int)&tx_buf[0];
	putmsg(0, &tx_buf[0], len);
}

/******************************************************************************
*  FUNCTION:    thread_info_msg                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Handles 'qfThreadInfo' and 'qsThreadInfo' query packet to get list of active 
*    threads.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     None
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void thread_info_msg(void)
{
	struct task_struct *thread = NULL;
	int threadid;
	threadref thref;

	char tx_buf[BUFMAX];
	char *pos = &tx_buf[0];
	int len = 0;
	int pkt_len = 0;
	int flag = 0;

	/* handle 'qsThreadInfo' packet */
	if(qsThreadInfo)	
	{		
		if(nThreadCount == PID_MAX_DEFAULT)
		{
			/* No more active threads are present so send 'l' to mark the end of 
			   thread list */
			putmsg(0, "l", 1);
			qsThreadInfo = 0;
			nThreadCount = 0;
			return;
		}

		threadid = nThreadCount;
	}
	else
	{
		/* This is 'qfThreadInfo' packet. GDB wont use PID 0, so start from 1 */
		threadid = 1;
		nThreadCount = 0;
	}


	/* Response from target starts with 'm' */
	*pos++ = 'm';
	pkt_len++;


	for (; threadid <= PID_MAX_DEFAULT; threadid++) 
	{
		read_lock(&tasklist_lock);
		thread = get_thread(threadid);
		read_unlock(&tasklist_lock);

		/* If it is a valid thread, add to packet */
		if (thread) 
		{
			int_to_threadref(&thref, threadid);
			pack_threadid(pos, &thref);
			pos += BUF_THREAD_ID_SIZE;
			*pos++ = ',';
			pkt_len += (BUF_THREAD_ID_SIZE +1);
		}

		/* Check if adding next thread id to packet will exceed the max. packet size.
		   If yes then send these many thread id's first, subsequent thread id's will
		   be sent when 'qsThreadInfo' packet is received from host */
		if( (pkt_len+BUF_THREAD_ID_SIZE) > (MAX_PACKET_LENGTH-1))
		{
			nThreadCount = threadid + 1;
			flag = 1;
			break;
		}
	}

	*--pos = 0;		/* Lose final comma */

	/* Check if all thread id's are traversed and included in the packet to be sent */
	if(threadid > PID_MAX_DEFAULT)
	{
		nThreadCount = PID_MAX_DEFAULT;
	}

	len = (int)pos - (int)&tx_buf[0];
	putmsg(0, &tx_buf[0], len);
}


/******************************************************************************
*  FUNCTION:    thread_info_msg                                  
*                                                                            
*  DESCRIPTION:                                                              
*    Handles 'ThreadExtraInfo' query packet to print some extra information about 
*    the thread. In this case we print the executable name.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     pkt : query packet from host.
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void thread_extra_info_msg(char* pkt)
{
	int threadid;
	struct task_struct *thread = NULL;
	char buffer[20], *ptr;
	int i, len;
	
	char tx_buf[BUFMAX];
	
	ptr = pkt;

	get_threadid_from_pkt(&ptr, &threadid);
	thread = get_thread(threadid);

	/* If we don't recognise it, say so */
	if (thread == NULL)
		strcpy(buffer, "(unknown)");
	else
		strcpy(buffer, thread->comm);

	/* Construct packet */
	for (i = 0, ptr = tx_buf; buffer[i]; i++)
		ptr = pack_hex_byte(ptr, buffer[i]);

#if	0 /* REVISIT! */
	if (thread->thread.pc == (unsigned long)ret_from_fork) {
		strcpy(buffer, "<new fork>");
		for (i = 0; buffer[i]; i++)
			ptr = pack_hex_byte(ptr, buffer[i]);
	}
#endif

*ptr = '\0';

	len = (int)ptr - (int)&tx_buf[0];
	putmsg(0, &tx_buf[0], len);
}


/******************************************************************************
*  FUNCTION:    query_msg                                  
*                                                                            
*  DESCRIPTION:                                                              
*     All query packets are received by this function and handled accordingly.
*                                                                            
*  INPUT PARAMETERS:                                                         
*     pkt : query packet from host.
*                                                                            
*  OUTPUT PARAMETERS:                                                        
*     None                                                        
*                                                                            
*  RETURN VALUES:                                                            
*                                                                            
*  NOTES:                                                                    
*                                                                            
******************************************************************************/
static void query_msg(char *pkt)
{
	char *q_start = pkt;

	/* qC = return current thread ID */
	if (strncmp(q_start, "C", 1) == 0)
		thread_id_msg();

	/* qfThreadInfo = query all threads (first) */
	else if (strncmp(q_start, "fThreadInfo", 11) == 0)
		thread_info_msg();

	/* qsThreadInfo = query all threads (subsequent). We know we have sent
	   them all after the qfThreadInfo message, so there are no to send */
	else if (strncmp(q_start, "sThreadInfo", 11) == 0)
	{
		qsThreadInfo = 1;
		thread_info_msg();		
	}

	/* qThreadExtraInfo = supply printable information per thread */
	else if (strncmp(q_start, "ThreadExtraInfo", 15) == 0)
		thread_extra_info_msg((q_start+strlen("ThreadExtraInfo")+1));
	
	/* Unsupported - empty message as per spec */
	else
		putmsg (0, "", 0);

}


#endif
