/*
 *  linux/drivers/mtd/cnxtnfc/generic.c
 *
 *
 * Based on linux/drivers/mtd/onenand/<>.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: generic.c 76911 2009-02-17 09:29:25Z balanid $
 *
 *  Overview:
 *   This is a device driver for Conexant's NAND Flash Controller(NFC).
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/platform_device.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/cnxtnfc.h>
#include <linux/mtd/partitions.h>

#include <asm/io.h>
#include <asm/mach/flash.h>

#define DRIVER_NAME	"cnxtnfc"

#ifdef CONFIG_MTD_PARTITIONS
static const char *part_probes[] = { "cmdlinepart", NULL  };
#endif

struct nfc_info {
	struct mtd_info       mtd;
	struct mtd_partition  *parts;
	struct nfc_dev        nfc;
};

static int __devinit generic_nfc_probe(struct platform_device *pdev) 
{
	struct nfc_info *info;

	struct flash_platform_data *pdata = pdev->dev.platform_data;
	int err;

   printk("generic_nfc_probe called\n");

	info = kzalloc(sizeof(struct nfc_info), GFP_KERNEL);
	if (!info)
		return -ENOMEM;
/* -->     No need for requesting a memory region and 
           no need to ioremap area for the base of the device.
           we are not doing a physical mapping for NAND devices <-- */

	info->nfc.irq = platform_get_irq(pdev, 0);

	info->mtd.name = pdev->dev.bus_id;
	info->mtd.priv = &info->nfc;
	info->mtd.owner = THIS_MODULE;

	if (nfc_scan(&info->mtd, 1)) {
		err = -ENXIO;
		goto out_iounmap;
	}

#ifdef CONFIG_MTD_PARTITIONS
	err = parse_mtd_partitions(&info->mtd, part_probes, &info->parts, 0);
	if (err > 0)
		add_mtd_partitions(&info->mtd, info->parts, err);
	else if (err < 0 && pdata->parts)
		add_mtd_partitions(&info->mtd, pdata->parts, pdata->nr_parts);
	else
#endif
		err = add_mtd_device(&info->mtd);

	dev_set_drvdata(&pdev->dev, info);

	return 0;

out_iounmap:
	kfree(info);

	return err;
}

static int __devexit generic_nfc_remove(struct platform_device *pdev)
{
	struct nfc_info *info = dev_get_drvdata(&pdev->dev);

	dev_set_drvdata(&pdev->dev, NULL);

	if (info) {
		if (info->parts)
			del_mtd_partitions(&info->mtd);
		else
			del_mtd_device(&info->mtd);

		nfc_release(&info->mtd);

		kfree(info);
	}

	return 0;
}

static struct mtd_partition cnxt_nfc_partitions[] = {
        {
                .name =          "nandpart",
                .size =          0xFF80000,
                .offset =        0x0,
        }
};

static struct flash_platform_data nfc_nand_parts = {
   .nr_parts = 1,
   .parts    = &cnxt_nfc_partitions[0]
};

static struct platform_driver generic_nfc_driver = {
	.probe		= generic_nfc_probe,
	.remove		= __devexit_p(generic_nfc_remove),
	.driver		= {
		.name	= DRIVER_NAME,
	},
};

static struct platform_device nfc_nand_flash = {
	.name		= DRIVER_NAME,
	.id		= 0,
	.dev     = {
      .platform_data = &nfc_nand_parts,
	},
	.num_resources = 0,
};

static int __init generic_nfc_init(void)
{
   int err;
   printk("generic_nfc_init called\n");
   err = platform_driver_register(&generic_nfc_driver);
   if (err == 0)
      err = platform_device_register(&nfc_nand_flash);
   return err;
}

static void __exit generic_nfc_exit(void)
{
   platform_device_unregister(&nfc_nand_flash);
   platform_driver_unregister(&generic_nfc_driver);
}

module_init(generic_nfc_init);
module_exit(generic_nfc_exit);

MODULE_ALIAS(DRIVER_NAME);
MODULE_LICENSE("GPL");
MODULE_AUTHOR("RadhaKumar Pulyala <radha.kumar@conexant.com>");
MODULE_DESCRIPTION("Glue layer for Conexant's NAND Flash Controller(NFC)");
