/*
 *  linux/drivers/mtd/cnxtnfc/nfc_base.c
 *
 *
 * Based on linux/drivers/mtd/onenand/<>.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  $Id: nfc_base.c 71368 2009-01-12 11:56:29Z dewangm $
 *
 *  Overview:
 *   This is a device driver for Conexant's NAND Flash Controller(NFC).
 */

/*#define CONFIG_MTD_DEBUG*/
/*#define CONFIG_MTD_DEBUG_VERBOSE 4*/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/jiffies.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <linux/dma-mapping.h>
#include <asm/io.h>
#include <linux/mtd/cnxtnfc.h>
#include "nfc_prv.h"

/**
 * nfc_oob_64 - oob info for large (2KB) page
 */
#if 0
static struct nand_ecclayout nfc_oob_64 = {
	.eccbytes	= 20,
	.eccpos		= {
		8, 9, 10, 11, 12,
		24, 25, 26, 27, 28,
		40, 41, 42, 43, 44,
		56, 57, 58, 59, 60,
		},
	.oobfree	= {
		{2, 3}, {14, 2}, {18, 3}, {30, 2},
		{34, 3}, {46, 2}, {50, 3}, {62, 2}
	}
};
#endif

static struct nand_ecclayout nfc_oob_64 = {
	.eccbytes	= 0,
	.oobfree	= {
		{2, 3}, {14, 2}, {18, 3}, {30, 2},
		{34, 3}, {46, 2}, {50, 3}, {62, 2}
	}
};

/**
 * nfc_oob_32 - oob info for middle (1KB) page
 */
static struct nand_ecclayout nfc_oob_32 = {
	.eccbytes	= 10,
	.eccpos		= {
		8, 9, 10, 11, 12,
		24, 25, 26, 27, 28,
		},
	.oobfree	= { {2, 3}, {14, 2}, {18, 3}, {30, 2} }
};

static const unsigned char ffchars[] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,	/* 16 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,	/* 32 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,	/* 48 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,	/* 64 */
};

/********************************************************/
extern int nfc_page_read (
               struct nfc_dev       *nfc, 
               unsigned char        *buffer, 
               int                  offset);

extern int nfc_page_write (
               struct nfc_dev       *nfc, 
               const unsigned char  *buffer, 
               int                  offset);

extern int nfc_block_erase( 
               struct nfc_dev       *nfc,
               int                  offset);

extern int nfc_spare_read(
               struct nfc_dev       *nfc, 
               unsigned char        *buffer, 
               int                  offset);

extern int nfc_spare_write(
               struct nfc_dev       *nfc, 
               const unsigned char  *buffer, 
               int                  offset);

extern int nfc_check_isbad( 
               struct nfc_dev       *nfc, 
               unsigned int         block_addr );

extern unsigned int nfc_read_id( 
               unsigned int         *maker_id, 
               unsigned int         *device_id );

extern unsigned int nfc_dev_reset(
               unsigned int         *status);

extern unsigned char nfc_wait_for_dev_ready(
               void);

extern int nfc_default_init(
               void);
/********************************************************/


inline static int nfc_page_address( struct nfc_dev *this, int offset)
{
   return (offset >> this->page_shift);
}

inline static int nfc_col_address( struct nfc_dev *this, int offset)
{
   return (offset & this->page_mask);
}

inline static int nfc_spare_col_address( struct nfc_dev *this, int offset)
{
   return ((offset & this->page_mask) | NFC_SPARE_COLUMN_MASK);
}

inline static int nfc_block_address( struct nfc_dev *this, int offset)
{
   return (offset >> this->erase_shift);
}

inline static int nfc_make_address( struct nfc_dev *this, int block, int page, int col)
{
   return ((block << this->erase_shift) | (page << this->page_shift) | (col & this->page_mask));
}

/* Use readl and writel for reading and writing registers in NFC space */

/**
 * nfc_wait - [DEFAULT] wait until the command is done
 * @param mtd		MTD device structure
 * @param state		state to select the max. timeout value
 *
 * Wait for command done. This applies to all NFC command
 * Read can take up to 30us, erase up to 2ms and program up to 350us
 * according to typical NAND Flash specs
 */
static int nfc_wait(struct nfc_dev *this, int state)
{
   /* RK: right now, we wait for the device to turn READY (on its R/B# signal) 
          No waiting for timeout.
   */
   return nfc_wait_for_dev_ready();
}

/* RK: In case of NFC, interrupt is only for ERROR control and debug.
       Will be taken care in future */

#if 0
/*
 * nfc_interrupt - [DEFAULT] nfc interrupt handler
 * @param irq		nfc interrupt number
 * @param dev_id	interrupt data
 *
 * complete the work
 */
static irqreturn_t nfc_interrupt(int irq, void *data)
{
	struct nfc_dev *this = (struct nfc_dev *) data;

	/* To handle shared interrupt */
	if (!this->complete.done)
		complete(&this->complete);

	return IRQ_HANDLED;
}

/*
 * nfc_interrupt_wait - [DEFAULT] wait until the command is done
 * @param mtd		MTD device structure
 * @param state		state to select the max. timeout value
 *
 * Wait for command done.
 */
static int nfc_interrupt_wait(struct nfc_dev *this, int state)
{
	wait_for_completion(&this->complete);
	return nfc_wait(this, state);
}

/*
 * nfc_try_interrupt_wait - [DEFAULT] try interrupt wait
 * @param mtd		MTD device structure
 * @param state		state to select the max. timeout value
 *
 * Try interrupt based wait (It is used one-time)
 */
static int nfc_try_interrupt_wait(struct mtd_info *mtd, int state)
{
	struct nfc_dev *this = mtd->priv;
	unsigned long remain, timeout;

	/* We use interrupt wait first */
	this->wait = nfc_interrupt_wait;

	timeout = msecs_to_jiffies(100);
	remain = wait_for_completion_timeout(&this->complete, timeout);
	if (!remain) {
		printk(KERN_INFO "NFC: There's no interrupt. "
				"We use the normal wait\n");

		/* Release the irq */
		free_irq(this->irq, this);

		this->wait = nfc_wait;
	}

	return nfc_wait(this, state);
}
#endif

/*
 * nfc_setup_wait - [NFC Interface] setup nfc wait method
 * @param mtd		MTD device structure
 *
 * There's two method to wait nfc work
 * 1. polling - read interrupt status register
 * 2. interrupt - use the kernel interrupt method
 */
static void nfc_setup_wait(struct mtd_info *mtd)
{
	struct nfc_dev *this = mtd->priv;

	init_completion(&this->complete);

	if (this->irq <= 0) {
		this->wait = nfc_wait;
		return;
	}
   /* RK: Setup interrupt for ERROR notification and debug ONLY */
}

/* RK: We donot have BufferRAM to operate on. */

/**
 * nfc_get_device - [GENERIC] Get chip for selected access
 * @param mtd		MTD device structure
 * @param new_state	the state which is requested
 *
 * Get the device and lock it for exclusive access
 */
static int nfc_get_device(struct mtd_info *mtd, int new_state)
{
	struct nfc_dev *this = mtd->priv;
	DECLARE_WAITQUEUE(wait, current);

	/*
	 * Grab the lock and see if the device is available
	 */
	while (1) {
		spin_lock(&this->chip_lock);
		if (this->state == FL_READY) {
			this->state = new_state;
			spin_unlock(&this->chip_lock);
			break;
		}
		if (new_state == FL_PM_SUSPENDED) {
			spin_unlock(&this->chip_lock);
			return (this->state == FL_PM_SUSPENDED) ? 0 : -EAGAIN;
		}
		set_current_state(TASK_UNINTERRUPTIBLE);
		add_wait_queue(&this->wq, &wait);
		spin_unlock(&this->chip_lock);
		schedule();
		remove_wait_queue(&this->wq, &wait);
	}

	return 0;
}

/**
 * nfc_release_device - [GENERIC] release chip
 * @param mtd		MTD device structure
 *
 * Deselect, release chip lock and wake up anyone waiting on the device
 */
static void nfc_release_device(struct mtd_info *mtd)
{
	struct nfc_dev *this = mtd->priv;

	/* Release the chip */
	spin_lock(&this->chip_lock);
	this->state = FL_READY;
	wake_up(&this->wq);
	spin_unlock(&this->chip_lock);
}

/**
 * nfc_read - [MTD Interface] Read data from flash
 * @param mtd		MTD device structure
 * @param from		offset to read from
 * @param len		number of bytes to read
 * @param retlen	pointer to variable to store the number of read bytes
 * @param buf		the databuffer to put data
 *
 * Read with ecc
*/
static int nfc_read(struct mtd_info *mtd, loff_t from, size_t len,
	size_t *retlen, u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	struct mtd_ecc_stats stats;
	unsigned int readlen = 0, column, page_addr;
   unsigned int nbytes;
	int thislen;
	int ret = 0, aligned = 1;

	DEBUG(MTD_DEBUG_LEVEL3, "nfc_read: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_read: Attempt read beyond end of device\n");
		*retlen = 0;
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nfc_get_device(mtd, FL_READING);

	/* TODO handling oob/ecc */

	stats = mtd->ecc_stats;

 	/* First, read the partial pages, if any */
   if (!this->page_buf)
   {
      DEBUG(MTD_DEBUG_LEVEL0, "nfc_read: nfc_dev->page_buf is NULL\n");
      *retlen = 0;
      return -EINVAL;
   }

   page_addr = from >> this->page_shift;
   column = from & this->page_mask;
   readlen = len;

   while(1)
   {
      nbytes = min(mtd->writesize - column, readlen);
      aligned = (nbytes == mtd->writesize);

      thislen = this->read_page(this, this->page_buf, page_addr);
      if (thislen <= 0)
      {
         DEBUG(MTD_DEBUG_LEVEL0, "nfc_read: this->read_page returned ERROR\n");
         *retlen = 0;
         return thislen;
      }
      if(!aligned)
      {
         memcpy(buf, this->page_buf+column, nbytes);
      }
      else
      {
         memcpy(buf, this->page_buf, nbytes);
      }
      buf += nbytes;
      readlen -= nbytes;

      if(readlen <= 0)
         break;

      column = 0;
      page_addr++;

      /* TODO: chip boundary */
   }

	/* Deselect and wake up anyone waiting on the device */
	nfc_release_device(mtd);

	/*
	 * TODO: Return success, if no ECC failures, else -EBADMSG
	 * fs driver will take care of that, because
	 * retlen == desired len and result == -EBADMSG
	 */
	(*retlen) = len - readlen;

	/* TODO: if (mtd->ecc_stats.failed - stats.failed) */
   if (0)
		return -EBADMSG;

	if (ret)
		return ret;

	return 0;
}

/**
 * nfc_do_read_oob - [MTD Interface] NFC read out-of-band
 * @param mtd		MTD device structure
 * @param from		offset to read from
 * @param len		number of bytes to read
 * @param retlen	pointer to variable to store the number of read bytes
 * @param buf		the databuffer to put data
 *
 * NFC read out-of-band data from the spare area
 */
int nfc_do_read_oob(struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	int read = 0, thislen, column;
	unsigned int page_addr;
	int ret = 0;

	DEBUG(MTD_DEBUG_LEVEL3, "nfc_read_oob: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);

	/* Initialize return length value */
	*retlen = 0;

	/* Do not allow reads past end of device */
	if (unlikely((from + len) > mtd->size)) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_read_oob: Attempt read beyond end of device\n");
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nfc_get_device(mtd, FL_READING);

	column = from & (mtd->oobsize - 1);

   page_addr = from >> this->page_shift;

	while (read < len) {
		cond_resched();

		thislen = mtd->oobsize - column;
		thislen = min_t(int, thislen, len);

      ret = this->read_spare(this, this->page_buf + mtd->writesize, page_addr);

		memcpy( buf, this->page_buf + mtd->writesize + column, thislen);

		if (ret) {
			DEBUG(MTD_DEBUG_LEVEL0, "nfc_read_oob: read failed = 0x%x\n", ret);
			goto out;
		}

		read += thislen;

		if (read == len)
			break;

		buf += thislen;

		/* Read more? */
		if (read < len) {
			/* Page size */
			from += mtd->writesize;
			column = 0;
		}
	}

out:
	/* Deselect and wake up anyone waiting on the device */
	nfc_release_device(mtd);

	*retlen = read;
	return ret;
}

/**
 * nfc_read_oob - [MTD Interface] NAND write data and/or out-of-band
 * @mtd:	MTD device structure
 * @from:	offset to read from
 * @ops:	oob operation description structure
 */
static int nfc_read_oob(struct mtd_info *mtd, loff_t from,
			    struct mtd_oob_ops *ops)
{
	BUG_ON(ops->mode != MTD_OOB_PLACE);

	return nfc_do_read_oob(mtd, from + ops->ooboffs, ops->ooblen,
				   &ops->oobretlen, ops->oobbuf);
}

#ifdef CONFIG_MTD_CNXTNFC_VERIFY_WRITE

/* TODO: these will be taken up later */

/**
 * nfc_verify_oob - [GENERIC] verify the oob contents after a write
 * @param mtd		MTD device structure
 * @param buf		the databuffer to verify
 * @param to		offset to read from
 * @param len		number of bytes to read and compare
 *
 */
static int nfc_verify_oob(struct mtd_info *mtd, const u_char *buf, loff_t to, int len)
{
	struct nfc_dev *this = mtd->priv;
	char *readp = this->page_buf;
	int column = to & (mtd->oobsize - 1);
	int status, i;

	this->command(mtd, CNXTNFC_CMD_READOOB, to, mtd->oobsize);
	nfc_update_bufferram(mtd, to, 0);
	status = this->wait(this, FL_READING);
	if (status)
		return status;

	this->read_bufferram(mtd, CNXTNFC_SPARERAM, readp, column, len);

	for(i = 0; i < len; i++)
		if (buf[i] != 0xFF && buf[i] != readp[i])
			return -EBADMSG;

	return 0;
}

/**
 * nfc_verify_page - [GENERIC] verify the chip contents after a write
 * @param mtd		MTD device structure
 * @param buf		the databuffer to verify
 *
 * Check DataRAM area directly
 */
static int nfc_verify_page(struct mtd_info *mtd, u_char *buf, loff_t addr)
{
	struct nfc_dev *this = mtd->priv;
	void __iomem *dataram0, *dataram1;
	int ret = 0;

	/* In partial page write, just skip it */
	if ((addr & (mtd->writesize - 1)) != 0)
		return 0;

	this->command(mtd, CNXTNFC_CMD_READ, addr, mtd->writesize);

	ret = this->wait(this, FL_READING);
	if (ret)
		return ret;

	nfc_update_bufferram(mtd, addr, 1);

	/* Check, if the two dataram areas are same */
	dataram0 = this->base + CNXTNFC_DATARAM;
	dataram1 = dataram0 + mtd->writesize;

	if (memcmp(dataram0, dataram1, mtd->writesize))
		return -EBADMSG;

	return 0;
}
#else
#define nfc_verify_page(...)	(0)
#define nfc_verify_oob(...)		(0)
#endif

#define NOTALIGNED(x)	((x & this->page_mask) != 0)

/**
 * nfc_write - [MTD Interface] write buffer to FLASH
 * @param mtd		MTD device structure
 * @param to		offset to write to
 * @param len		number of bytes to write
 * @param retlen	pointer to variable to store the number of written bytes
 * @param buf		the data to write
 *
 * Write with ECC
 */
static int nfc_write(struct mtd_info *mtd, loff_t to, size_t len,
	size_t *retlen, const u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	int written = 0;
	int ret = 0;
	int column, subpage;
   unsigned int page_addr;

	DEBUG(MTD_DEBUG_LEVEL3, "nfc_write: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Initialize retlen, in case of early exit */
	*retlen = 0;

	/* Do not allow writes past end of device */
	if (unlikely((to + len) > mtd->size)) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_write: Attempt write to past end of device\n");
		return -EINVAL;
	}

	/* Reject writes, which are not page aligned */
   if (unlikely(NOTALIGNED(to)) || unlikely(NOTALIGNED(len))) {
          DEBUG(MTD_DEBUG_LEVEL0, "nfc_write: Attempt to write not page aligned data\n");
          return -EINVAL;
   }

	column = to & (this->page_mask);
	subpage = column || (len & (this->page_mask));
   page_addr = to >> this->page_shift;
   
	/* Grab the lock and see if the device is available */
	nfc_get_device(mtd, FL_WRITING);

	/* Loop until all data write */
	while (written < len) {
		int bytes = mtd->writesize;
		int thislen = min_t(int, bytes, len - written);
		u_char *wbuf = (u_char *) buf;

		cond_resched();

      column = to & (this->page_mask);
      subpage = column || (len & (this->page_mask));
      page_addr = to >> this->page_shift;

		/* Partial page write */
		if (subpage) {
			bytes = min_t(int, bytes - column, (int) len);
			memset(this->page_buf, 0xff, mtd->writesize);
			wbuf = this->page_buf;
			/* Even though partial write, we need page size */
			thislen = mtd->writesize;
		}

		memcpy(this->page_buf + column, buf, bytes);
		memcpy(this->page_buf + mtd->writesize, ffchars, mtd->oobsize);

		ret = this->write_page(this, this->page_buf, page_addr);

		/* TODO: ret = this->wait(this, FL_WRITING); */
		if (ret) {
			DEBUG(MTD_DEBUG_LEVEL0, "nfc_write: write failed %d\n", ret);
			break;
		}

		/* Only check verify write turn on */
		ret = nfc_verify_page(mtd, (u_char *) wbuf, page_addr);
		if (ret) {
			DEBUG(MTD_DEBUG_LEVEL0, "nfc_write: verify failed %d\n", ret);
			break;
		}

		written += thislen;

		if (written == len)
			break;

		column = 0;
		to += thislen;
		buf += thislen;
	}

	/* Deselect and wake up anyone waiting on the device */
	nfc_release_device(mtd);

	*retlen = written;
	return ret;
}

/**
 * nfc_do_write_oob - [Internal] NFC write out-of-band
 * @param mtd		MTD device structure
 * @param to		offset to write to
 * @param len		number of bytes to write
 * @param retlen	pointer to variable to store the number of written bytes
 * @param buf		the data to write
 *
 * Cnxt NFC write out-of-band
 */
static int nfc_do_write_oob(struct mtd_info *mtd, loff_t to, size_t len,
				size_t *retlen, const u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	int column, ret = 0;
	int written = 0;
   unsigned int page_addr;
   
	DEBUG(MTD_DEBUG_LEVEL3, "nfc_write_oob: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Initialize retlen, in case of early exit */
	*retlen = 0;

	/* Do not allow writes past end of device */
	if (unlikely((to + len) > mtd->size)) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_write_oob: Attempt write to past end of device\n");
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nfc_get_device(mtd, FL_WRITING);

   page_addr = to >> this->page_shift;

	/* Loop until all data write */
	while (written < len) {
		int thislen = min_t(int, mtd->oobsize, len - written);

		cond_resched();

		column = to & (mtd->oobsize - 1);

		/* We send data to spare ram with oobsize
		 * to prevent byte access */
		memset(this->page_buf + mtd->writesize, 0xff, mtd->oobsize);
		memcpy(this->page_buf + mtd->writesize + column, buf, thislen);

		ret = this->write_spare(this, this->page_buf + mtd->writesize, page_addr);

		/* ret = this->wait(this, FL_WRITING); */
		if (ret) {
			DEBUG(MTD_DEBUG_LEVEL0, "nfc_write_oob: write filaed %d\n", ret);
			goto out;
		}

		ret = nfc_verify_oob(mtd, buf, page_addr, thislen);
		if (ret) {
			DEBUG(MTD_DEBUG_LEVEL0, "nfc_write_oob: verify failed %d\n", ret);
			goto out;
		}

		written += thislen;

		if (written == len)
			break;

		to += thislen;
		buf += thislen;
	}

out:
	/* Deselect and wake up anyone waiting on the device */
	nfc_release_device(mtd);

	*retlen = written;
	return ret;
}

/**
 * nfc_write_oob - [MTD Interface] NAND write data and/or out-of-band
 * @mtd:	MTD device structure
 * @from:	offset to read from
 * @ops:	oob operation description structure
 */
static int nfc_write_oob(struct mtd_info *mtd, loff_t to,
			     struct mtd_oob_ops *ops)
{
	BUG_ON(ops->mode != MTD_OOB_PLACE);

	return nfc_do_write_oob(mtd, to + ops->ooboffs, ops->ooblen,
				    &ops->oobretlen, ops->oobbuf);
}

/**
 * nfc_block_checkbad - [GENERIC] Check if a block is marked bad
 * @param mtd		MTD device structure
 * @param ofs		offset from device start
 * @param getchip 0, if the chip is already selected
 * @param allowbbt   1, if its allowed to access the bbt area
 *
 * Check, if the block is bad. Either by reading the bad block table or
 * calling of the scan function.
 */
static int nfc_block_checkbad(struct mtd_info *mtd, loff_t ofs, int getchip, int allowbbt)
{
	struct nfc_dev *this = mtd->priv;
	struct bbm_info *bbm = this->bbm;

	/* Return info from the table */
	return bbm->isbad_bbt(mtd, ofs, allowbbt);
}

/**
 * nfc_erase - [MTD Interface] erase block(s)
 * @param mtd		MTD device structure
 * @param instr		erase instruction
 *
 * Erase one or more blocks
 */
static int nfc_erase(struct mtd_info *mtd, struct erase_info *instr)
{
	struct nfc_dev *this = mtd->priv;
	unsigned int block_size;
	loff_t addr;
	int len;
	int ret = 0;

	DEBUG(MTD_DEBUG_LEVEL3, "nfc_erase: start = 0x%08x, len = %i\n", (unsigned int) instr->addr, (unsigned int) instr->len);

	block_size = (1 << this->erase_shift);

	/* Start address must align on block boundary */
	if (unlikely(instr->addr & (block_size - 1))) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_erase: Unaligned address\n");
		return -EINVAL;
	}

	/* Length must align on block boundary */
	if (unlikely(instr->len & (block_size - 1))) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_erase: Length not block aligned\n");
		return -EINVAL;
	}

	/* Do not allow erase past end of device */
	if (unlikely((instr->len + instr->addr) > mtd->size)) {
		DEBUG(MTD_DEBUG_LEVEL0, "nfc_erase: Erase past end of device\n");
		return -EINVAL;
	}

	instr->fail_addr = 0xffffffff;

	/* Grab the lock and see if the device is available */
	nfc_get_device(mtd, FL_ERASING);

	/* Loop throught the pages */
	len = instr->len;
	addr = instr->addr;

	instr->state = MTD_ERASING;

	while (len) {
		cond_resched();

		/* Check if we have a bad block, we do not erase bad blocks */
		if (nfc_block_checkbad(mtd, addr, 0, 0)) {
			printk (KERN_WARNING "nfc_erase: attempt to erase a bad block at addr 0x%08x\n", (unsigned int) addr);
			instr->state = MTD_ERASE_FAILED;
			goto erase_exit;
		}

      /* dewangm */
		ret = this->erase_block(this,
	         (addr >> this->page_shift  /*this->erase_shift*/)); /* Block address */

      /* TODO: ret = this->wait(mtd, FL_ERASING); */

		/* Check, if it is write protected */
		if (ret) {
			DEBUG(MTD_DEBUG_LEVEL0, "nfc_erase: Failed erase, block %d\n", (unsigned) (addr >> this->erase_shift));
			instr->state = MTD_ERASE_FAILED;
			instr->fail_addr = addr;
			goto erase_exit;
		}

		len -= block_size;
		addr += block_size;
	}

	instr->state = MTD_ERASE_DONE;

erase_exit:

	ret = instr->state == MTD_ERASE_DONE ? 0 : -EIO;
	/* Do call back function */
	if (!ret)
		mtd_erase_callback(instr);

	/* Deselect and wake up anyone waiting on the device */
	nfc_release_device(mtd);

	return ret;
}

/**
 * nfc_sync - [MTD Interface] sync
 * @param mtd		MTD device structure
 *
 * Sync is actually a wait for chip ready function
 */
static void nfc_sync(struct mtd_info *mtd)
{
	DEBUG(MTD_DEBUG_LEVEL3, "nfc_sync: called\n");

	/* Grab the lock and see if the device is available */
	nfc_get_device(mtd, FL_SYNCING);

	/* Release it and go back */
	nfc_release_device(mtd);
}

/**
 * nfc_block_isbad - [MTD Interface] Check whether the block at the given offset is bad
 * @param mtd		MTD device structure
 * @param ofs		offset relative to mtd start
 *
 * Check whether the block is bad
 */
static int nfc_block_isbad(struct mtd_info *mtd, loff_t ofs)
{
	/* Check for invalid offset */
	if (ofs > mtd->size)
		return -EINVAL;

	return nfc_block_checkbad(mtd, ofs, 1, 0);
}

/**
 * nfc_default_block_markbad - [DEFAULT] mark a block bad
 * @param mtd		MTD device structure
 * @param ofs		offset from device start
 *
 * This is the default implementation, which can be overridden by
 * a hardware specific driver.
 */
static int nfc_default_block_markbad(struct mtd_info *mtd, loff_t ofs)
{
	struct nfc_dev *this = mtd->priv;
	struct bbm_info *bbm = this->bbm;
	u_char buf[2] = {0, 0};
	size_t retlen;
	int block;

	/* Get block number */
	block = ((int) ofs) >> bbm->bbt_erase_shift;
        if (bbm->bbt)
                bbm->bbt[block >> 2] |= 0x01 << ((block & 0x03) << 1);

        /* We write two bytes, so we dont have to mess with 16 bit access */
        ofs += mtd->oobsize + (bbm->badblockpos & ~0x01);
        return nfc_do_write_oob(mtd, ofs , 2, &retlen, buf);
}

/**
 * nfc_block_markbad - [MTD Interface] Mark the block at the given offset as bad
 * @param mtd		MTD device structure
 * @param ofs		offset relative to mtd start
 *
 * Mark the block as bad
 */
static int nfc_block_markbad(struct mtd_info *mtd, loff_t ofs)
{
	struct nfc_dev *this = mtd->priv;
	int ret;

	ret = nfc_block_isbad(mtd, ofs);
	if (ret) {
		/* If it was bad already, return success and do nothing */
		if (ret > 0)
			return 0;
		return ret;
	}

	return this->block_markbad(mtd, ofs);
}

#if 0
/**
 * nfc_do_lock_cmd - [NFC Interface] Lock or unlock block(s)
 * @param mtd		MTD device structure
 * @param ofs		offset relative to mtd start
 * @param len		number of bytes to lock or unlock
 *
 * Lock or unlock one or more blocks
 */
static int nfc_do_lock_cmd(struct mtd_info *mtd, loff_t ofs, size_t len, int cmd)
{
	return 0;
}

/**
 * nfc_lock - [MTD Interface] Lock block(s)
 * @param mtd		MTD device structure
 * @param ofs		offset relative to mtd start
 * @param len		number of bytes to lock
 *
 * Lock one or more blocks
 */
static int nfc_lock(struct mtd_info *mtd, loff_t ofs, size_t len)
{
	return 0; /*nfc_do_lock_cmd(mtd, ofs, len, CNXTNFC_CMD_LOCK);*/
}

/**
 * nfc_unlock - [MTD Interface] Unlock block(s)
 * @param mtd		MTD device structure
 * @param ofs		offset relative to mtd start
 * @param len		number of bytes to unlock
 *
 * Unlock one or more blocks
 */
static int nfc_unlock(struct mtd_info *mtd, loff_t ofs, size_t len)
{
	return 0; /*nfc_do_lock_cmd(mtd, ofs, len, CNXTNFC_CMD_UNLOCK);*/
}

/**
 * nfc_check_lock_status - [NFC Interface] Check lock status
 * @param this		nfc chip data structure
 *
 * Check lock status
 */
static void nfc_check_lock_status(struct nfc_dev *this)
{
}

/**
 * nfc_unlock_all - [NFC Interface] unlock all blocks
 * @param mtd		MTD device structure
 *
 * Unlock all blocks
 */
static int nfc_unlock_all(struct mtd_info *mtd)
{
	return 0;
}
#endif

#ifdef CONFIG_MTD_CNXTNFC_OTP

/* TODO: OTP */

/* Interal OTP operation */
typedef int (*otp_op_t)(struct mtd_info *mtd, loff_t form, size_t len,
		size_t *retlen, u_char *buf);

/**
 * do_otp_read - [DEFAULT] Read OTP block area
 * @param mtd		MTD device structure
 * @param from		The offset to read
 * @param len		number of bytes to read
 * @param retlen	pointer to variable to store the number of readbytes
 * @param buf		the databuffer to put/get data
 *
 * Read OTP block area.
 */
static int do_otp_read(struct mtd_info *mtd, loff_t from, size_t len,
		size_t *retlen, u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	int ret;

	/* Enter OTP access mode */
	this->command(mtd, CNXTNFC_CMD_OTP_ACCESS, 0, 0);
	this->wait(this, FL_OTPING);

	ret = mtd->read(mtd, from, len, retlen, buf);

	/* Exit OTP access mode */
	this->command(mtd, CNXTNFC_CMD_RESET, 0, 0);
	this->wait(this, FL_RESETING);

	return ret;
}

/**
 * do_otp_write - [DEFAULT] Write OTP block area
 * @param mtd		MTD device structure
 * @param from		The offset to write
 * @param len		number of bytes to write
 * @param retlen	pointer to variable to store the number of write bytes
 * @param buf		the databuffer to put/get data
 *
 * Write OTP block area.
 */
static int do_otp_write(struct mtd_info *mtd, loff_t from, size_t len,
		size_t *retlen, u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	unsigned char *pbuf = buf;
	int ret;

	/* Force buffer page aligned */
	if (len < mtd->writesize) {
		memcpy(this->page_buf, buf, len);
		memset(this->page_buf + len, 0xff, mtd->writesize - len);
		pbuf = this->page_buf;
		len = mtd->writesize;
	}

	/* Enter OTP access mode */
	this->command(mtd, CNXTNFC_CMD_OTP_ACCESS, 0, 0);
	this->wait(this, FL_OTPING);

	ret = mtd->write(mtd, from, len, retlen, pbuf);

	/* Exit OTP access mode */
	this->command(mtd, CNXTNFC_CMD_RESET, 0, 0);
	this->wait(this, FL_RESETING);

	return ret;
}

/**
 * do_otp_lock - [DEFAULT] Lock OTP block area
 * @param mtd		MTD device structure
 * @param from		The offset to lock
 * @param len		number of bytes to lock
 * @param retlen	pointer to variable to store the number of lock bytes
 * @param buf		the databuffer to put/get data
 *
 * Lock OTP block area.
 */
static int do_otp_lock(struct mtd_info *mtd, loff_t from, size_t len,
		size_t *retlen, u_char *buf)
{
	struct nfc_dev *this = mtd->priv;
	int ret;

	/* Enter OTP access mode */
	this->command(mtd, CNXTNFC_CMD_OTP_ACCESS, 0, 0);
	this->wait(this, FL_OTPING);

	ret = nfc_do_write_oob(mtd, from, len, retlen, buf);

	/* Exit OTP access mode */
	this->command(mtd, CNXTNFC_CMD_RESET, 0, 0);
	this->wait(this, FL_RESETING);

	return ret;
}

/**
 * nfc_otp_walk - [DEFAULT] Handle OTP operation
 * @param mtd		MTD device structure
 * @param from		The offset to read/write
 * @param len		number of bytes to read/write
 * @param retlen	pointer to variable to store the number of read bytes
 * @param buf		the databuffer to put/get data
 * @param action	do given action
 * @param mode		specify user and factory
 *
 * Handle OTP operation.
 */
static int nfc_otp_walk(struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf,
			otp_op_t action, int mode)
{
	struct nfc_dev *this = mtd->priv;
	int otp_pages;
	int density;
	int ret = 0;

	*retlen = 0;

	density = this->device_id >> CNXTNFC_DEVICE_DENSITY_SHIFT;
	if (density < CNXTNFC_DEVICE_DENSITY_512Mb)
		otp_pages = 20;
	else
		otp_pages = 10;

	if (mode == MTD_OTP_FACTORY) {
		from += mtd->writesize * otp_pages;
		otp_pages = 64 - otp_pages;
	}

	/* Check User/Factory boundary */
	if (((mtd->writesize * otp_pages) - (from + len)) < 0)
		return 0;

	while (len > 0 && otp_pages > 0) {
		if (!action) {	/* OTP Info functions */
			struct otp_info *otpinfo;

			len -= sizeof(struct otp_info);
			if (len <= 0)
				return -ENOSPC;

			otpinfo = (struct otp_info *) buf;
			otpinfo->start = from;
			otpinfo->length = mtd->writesize;
			otpinfo->locked = 0;

			from += mtd->writesize;
			buf += sizeof(struct otp_info);
			*retlen += sizeof(struct otp_info);
		} else {
			size_t tmp_retlen;
			int size = len;

			ret = action(mtd, from, len, &tmp_retlen, buf);

			buf += size;
			len -= size;
			*retlen += size;

			if (ret < 0)
				return ret;
		}
		otp_pages--;
	}

	return 0;
}

/**
 * nfc_get_fact_prot_info - [MTD Interface] Read factory OTP info
 * @param mtd		MTD device structure
 * @param buf		the databuffer to put/get data
 * @param len		number of bytes to read
 *
 * Read factory OTP info.
 */
static int nfc_get_fact_prot_info(struct mtd_info *mtd,
			struct otp_info *buf, size_t len)
{
	size_t retlen;
	int ret;

	ret = nfc_otp_walk(mtd, 0, len, &retlen, (u_char *) buf, NULL, MTD_OTP_FACTORY);

	return ret ? : retlen;
}

/**
 * nfc_read_fact_prot_reg - [MTD Interface] Read factory OTP area
 * @param mtd		MTD device structure
 * @param from		The offset to read
 * @param len		number of bytes to read
 * @param retlen	pointer to variable to store the number of read bytes
 * @param buf		the databuffer to put/get data
 *
 * Read factory OTP area.
 */
static int nfc_read_fact_prot_reg(struct mtd_info *mtd, loff_t from,
			size_t len, size_t *retlen, u_char *buf)
{
	return nfc_otp_walk(mtd, from, len, retlen, buf, do_otp_read, MTD_OTP_FACTORY);
}

/**
 * nfc_get_user_prot_info - [MTD Interface] Read user OTP info
 * @param mtd		MTD device structure
 * @param buf		the databuffer to put/get data
 * @param len		number of bytes to read
 *
 * Read user OTP info.
 */
static int nfc_get_user_prot_info(struct mtd_info *mtd,
			struct otp_info *buf, size_t len)
{
	size_t retlen;
	int ret;

	ret = nfc_otp_walk(mtd, 0, len, &retlen, (u_char *) buf, NULL, MTD_OTP_USER);

	return ret ? : retlen;
}

/**
 * nfc_read_user_prot_reg - [MTD Interface] Read user OTP area
 * @param mtd		MTD device structure
 * @param from		The offset to read
 * @param len		number of bytes to read
 * @param retlen	pointer to variable to store the number of read bytes
 * @param buf		the databuffer to put/get data
 *
 * Read user OTP area.
 */
static int nfc_read_user_prot_reg(struct mtd_info *mtd, loff_t from,
			size_t len, size_t *retlen, u_char *buf)
{
	return nfc_otp_walk(mtd, from, len, retlen, buf, do_otp_read, MTD_OTP_USER);
}

/**
 * nfc_write_user_prot_reg - [MTD Interface] Write user OTP area
 * @param mtd		MTD device structure
 * @param from		The offset to write
 * @param len		number of bytes to write
 * @param retlen	pointer to variable to store the number of write bytes
 * @param buf		the databuffer to put/get data
 *
 * Write user OTP area.
 */
static int nfc_write_user_prot_reg(struct mtd_info *mtd, loff_t from,
			size_t len, size_t *retlen, u_char *buf)
{
	return nfc_otp_walk(mtd, from, len, retlen, buf, do_otp_write, MTD_OTP_USER);
}

/**
 * nfc_lock_user_prot_reg - [MTD Interface] Lock user OTP area
 * @param mtd		MTD device structure
 * @param from		The offset to lock
 * @param len		number of bytes to unlock
 *
 * Write lock mark on spare area in page 0 in OTP block
 */
static int nfc_lock_user_prot_reg(struct mtd_info *mtd, loff_t from,
			size_t len)
{
	unsigned char oob_buf[64];
	size_t retlen;
	int ret;

	memset(oob_buf, 0xff, mtd->oobsize);
	/*
	 * Note: OTP lock operation
	 *       OTP block : 0xXXFC
	 *       1st block : 0xXXF3 (If chip support)
	 *       Both      : 0xXXF0 (If chip support)
	 */
	oob_buf[CNXTNFC_OTP_LOCK_OFFSET] = 0xFC;

	/*
	 * Write lock mark to 8th word of sector0 of page0 of the spare0.
	 * We write 16 bytes spare area instead of 2 bytes.
	 */
	from = 0;
	len = 16;

	ret = nfc_otp_walk(mtd, from, len, &retlen, oob_buf, do_otp_lock, MTD_OTP_USER);

	return ret ? : retlen;
}
#endif	/* CONFIG_MTD_CNXTNFC_OTP */

/**
 * nfc_lock_scheme - Check and set NFC lock scheme
 * @param mtd		MTD data structure
 *
 * Check and set NFC lock scheme
 */
static void nfc_lock_scheme(struct mtd_info *mtd)
{
   /* TODO: Lock scheme update as part of low level driver integration */
}

/**
 * nfc_print_device_info - Print device ID
 * @param device        device ID
 *
 * Print device ID
 */
static void nfc_print_device_info(int mfr_id, int dev_id)
{
  printk(KERN_INFO "NAND Flash manufacturer ID: %d, device ID: %d\n", mfr_id, dev_id);
}

/*
*	Chip ID list
*
*  Name, manufacturer ID, dev ID, pagesize in bytes, spare size in bytes, 
*           chipsize in MegaByte, eraseblock size in bytes, options
*
*	Pagesize; 0, 512, 2048
*	0	get this information from the extended chip ID
+	256	256 Byte page size
*	512	512 Byte page size
*/
struct nand_flash_dev nand_flash_ids[] = {
	{"NAND 256MiB 3V 8-bit", NAND_MFR_MICRON, 0xDA, 2048, 64, 0x10000000, 0x20000, 0},
	{"NAND 256MiB 1.8V 8-bit", NAND_MFR_MICRON, 0xAA, 2048, 64, 0x10000000, 0x20000, 0},
	{"NAND 256MiB 2.7V 8-bit", NAND_MFR_SAMSUNG, 0xDA, 2048, 64, 0x10000000, 0x20000, 0},
	{NULL,0,0,0,0,0,0}
};

struct nand_flash_dev* nand_dev_get(int mfr_id, int dev_id)
{
   int size = ARRAY_SIZE(nand_flash_ids);
   int i;

   for ( i = 0; i < size; i++ )
      if ( mfr_id == nand_flash_ids[i].mfr_id && dev_id == nand_flash_ids[i].dev_id )
         break;
   if ( i < size )
      return &nand_flash_ids[i];

   return NULL;
}


/**
 * nfc_check_maf - Check manufacturer ID
 * @param manuf         manufacturer ID
 *
 * Check manufacturer ID
 */
static int nfc_check_maf(int manuf)
{
	int size = ARRAY_SIZE(nand_flash_ids);
	char *name;
        int i;

	for (i = 0; i < size; i++)
      if (manuf == nand_flash_ids[i].mfr_id)
         break;

	if (i < size)
		name = nand_flash_ids[i].name;
	else
		name = "Unknown";

	printk(KERN_DEBUG "%s (0x%0x)\n", name, manuf);

	return (i == size);
}

/**
 * nfc_probe - [NFC Interface] Probe the CnxtNFC device
 * @param mtd		MTD device structure
 *
 * NFC detection method:
 *   Compare the the values from command with ones from register
 */
static int nfc_probe(struct mtd_info *mtd)
{
	struct nfc_dev *this = mtd->priv;
	int mfr_id = 0, dev_id = 0;
   struct nand_flash_dev* nanddev;

   nfc_default_init();
	/* Read manufacturer and device IDs from BootRAM */
   nfc_read_id(&mfr_id, &dev_id);

   if (mfr_id == 0 || dev_id == 0 )
      return -ENXIO;

	/* Check manufacturer ID */
	if (nfc_check_maf(mfr_id))
   {  
      printk("NFC: Unsupported NAND flash manufacturer.\n");
		return -ENXIO;
   }

   nanddev = nand_dev_get(mfr_id, dev_id);

   if (!nanddev)
   {  
      printk("NFC: Unsupported NAND flash device.\n");
      return -ENXIO;
   }

	/* Reset NAND device(s) */
	nfc_dev_reset(NULL);

	/* Wait reset */
	this->wait(this, FL_RESETING);

	/* Flash device information */
	nfc_print_device_info(mfr_id, dev_id);
	this->mfr_id    = mfr_id;
	this->device_id = dev_id;

	this->flashsize = nanddev->chipsize;

	/* NAND page size & block size */
	/* The data buffer size is equal to page size */
	mtd->writesize = nanddev->pagesize;
	mtd->oobsize = nanddev->sparesize;
	mtd->erasesize = nanddev->erasesize;

	this->erase_shift = ffs(mtd->erasesize) - 1;
	this->page_shift = ffs(mtd->writesize) - 1;
	this->ppb_shift = (this->erase_shift - this->page_shift);
	this->page_mask = mtd->writesize - 1;

	/* REVIST: Multichip handling */

	mtd->size = this->flashsize;

	/* Check NFC lock scheme */
	nfc_lock_scheme(mtd);

	return 0;
}

/**
 * nfc_suspend - [MTD Interface] Suspend the NFC flash
 * @param mtd		MTD device structure
 */
static int nfc_suspend(struct mtd_info *mtd)
{
	return nfc_get_device(mtd, FL_PM_SUSPENDED);
}

/**
 * nfc_resume - [MTD Interface] Resume the NFC flash
 * @param mtd		MTD device structure
 */
static void nfc_resume(struct mtd_info *mtd)
{
	struct nfc_dev *this = mtd->priv;

	if (this->state == FL_PM_SUSPENDED)
		nfc_release_device(mtd);
	else
		printk(KERN_ERR "resume() called for the chip which is not"
				"in suspended state\n");
}

/**
 * nfc_scan - [NFC Interface] Scan for NAND device(s)
 * @param mtd		MTD device structure
 * @param maxchips	Number of chips to scan for
 *
 * This fills out all the not initialized function pointers
 * with the defaults.
 * The flash ID is read and the mtd/chip structures are
 * filled with the appropriate values.
 */
int nfc_scan(struct mtd_info *mtd, int maxchips)
{
	struct nfc_dev *this = mtd->priv;

	if (!this->wait)
		nfc_setup_wait(mtd);

	if (!this->block_markbad)
		this->block_markbad = nfc_default_block_markbad;
	if (!this->scan_bbt)
		this->scan_bbt = nfc_default_bbt;

	if (!this->read_page)
		this->read_page   = nfc_page_read;
	if (!this->write_page)
		this->write_page  = nfc_page_write;
	if (!this->read_spare)
		this->read_spare  = nfc_spare_read;
	if (!this->write_spare)
		this->write_spare = nfc_spare_write;
	if (!this->check_isbad)
		this->check_isbad = nfc_check_isbad;
	if (!this->erase_block)
		this->erase_block = nfc_block_erase;

	if (nfc_probe(mtd))
		return -ENXIO;

	/* Allocate buffers, if necessary */
	if (!this->page_buf) {
		size_t len;
		len = mtd->writesize + mtd->oobsize;
		if (NFC_DMA_ALLOC_COHERENT == YES)
		{
         this->page_buf = dma_alloc_coherent(NULL, len, &this->dmaaddr, GFP_KERNEL);
		}
		else
		{
   		this->page_buf = kmalloc(len, GFP_KERNEL);
      }
		if (!this->page_buf) {
			printk(KERN_ERR "nfc_scan(): Can't allocate page_buf\n");
			return -ENOMEM;
		}
	}

	this->state = FL_READY;
	init_waitqueue_head(&this->wq);
	spin_lock_init(&this->chip_lock);

	/*
	 * Allow subpage writes up to oobsize.
	 */
	switch (mtd->oobsize) {
	case 64:
		this->ecclayout = &nfc_oob_64;
		mtd->subpage_sft = 2;
		break;

	case 32:
		this->ecclayout = &nfc_oob_32;
		mtd->subpage_sft = 1;
		break;

   /* RK: Add case for oobsize = 16 = SLC small */
	default:
		printk(KERN_WARNING "No OOB scheme defined for oobsize %d\n",
			mtd->oobsize);
		mtd->subpage_sft = 0;
		/* To prevent kernel oops */
		this->ecclayout = &nfc_oob_32;
		break;
	}

	this->subpagesize = mtd->writesize >> mtd->subpage_sft;
	mtd->ecclayout = this->ecclayout;

	/* Fill in remaining MTD driver data */
	mtd->type = MTD_NANDFLASH;
	mtd->flags = MTD_CAP_NANDFLASH;
	mtd->ecctype = MTD_ECC_SW;
	mtd->erase = nfc_erase;
	mtd->point = NULL;
	mtd->unpoint = NULL;
	mtd->read = nfc_read;
	mtd->write = nfc_write;
	mtd->read_oob = nfc_read_oob;
	mtd->write_oob = nfc_write_oob;
   /* RK: OTP will be supported a little later. 
      Once the NFC support is complete */
#ifdef CONFIG_MTD_CNXTNFC_OTP
	mtd->get_fact_prot_info = nfc_get_fact_prot_info;
	mtd->read_fact_prot_reg = nfc_read_fact_prot_reg;
	mtd->get_user_prot_info = nfc_get_user_prot_info;
	mtd->read_user_prot_reg = nfc_read_user_prot_reg;
	mtd->write_user_prot_reg = nfc_write_user_prot_reg;
	mtd->lock_user_prot_reg = nfc_lock_user_prot_reg;
#endif
	mtd->sync = nfc_sync;
	mtd->lock = NULL;
	mtd->unlock = NULL;
	mtd->suspend = nfc_suspend;
	mtd->resume = nfc_resume;
	mtd->block_isbad = nfc_block_isbad;
	mtd->block_markbad = nfc_block_markbad;
	mtd->owner = THIS_MODULE;

	return this->scan_bbt(mtd);
}

/**
 * nfc_release - [NFC Interface] Free resources held by the CnxtNFC device
 * @param mtd		MTD device structure
 */
void nfc_release(struct mtd_info *mtd)
{
	struct nfc_dev *this = mtd->priv;

#ifdef CONFIG_MTD_PARTITIONS
	/* Deregister partitions */
	del_mtd_partitions (mtd);
#endif
	/* Deregister the device */
	del_mtd_device (mtd);

	/* Free bad block table memory, if allocated */
	if (this->bbm)
		kfree(this->bbm);
	/* Buffer allocated by nfc_scan */

	if (this->page_buf)
   {  
      if (NFC_DMA_ALLOC_COHERENT)
      {
   		size_t len;
   		len = mtd->writesize + mtd->oobsize;
   		dma_free_coherent(NULL, len, this->page_buf, this->dmaaddr);
		}
		else
		{
         kfree(this->page_buf);
	   }
      this->page_buf = NULL;
   }
}

EXPORT_SYMBOL_GPL(nfc_scan);
EXPORT_SYMBOL_GPL(nfc_release);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Kyungmin Park <kyungmin.park@samsung.com>");
MODULE_DESCRIPTION("Generic NFC flash driver code");
