/******************************************************************************/
/*                   CONEXANT PROPRIETARY AND CONFIDENTIAL                    */
/*                        SOFTWARE FILE/MODULE HEADER                         */
/*                 Copyright Conexant Systems Inc. 2007                       */
/*                            All Rights Reserved                             */
/******************************************************************************/
/*
 * Filename:        nfc_prv.c
 *
 *
 * Description:     API implementation for NAND Flash Controller (NFC) driver.
 *
 *
 * Author:          RadhaKumar Pulyala
 *
 ******************************************************************************/
/* $Id: nfc_prv.c 101070 2009-07-02 06:02:38Z rathd $
 ******************************************************************************/
/*****************/
/* Include Files */
/*****************/

/******************************************************************************
 * IMPORTANT
 * ---------
 *
 * PLEASE ENSURE THAT EACH TIME ANY OF THE FOLLOWING FILES ARE MODIFIED, 
 * THE CHANGES ARE COMMITTED TO THE OTHER FILE(S) ALSO.
 *
 * - target/src/linux/kernel/linux-2.6.18.5/drivers/mtd/cnxtnfc/nfc_prv.c
 * - target/src/u-boot/drivers/mtd/nfc/nfc_prv.c
 *
 *****************************************************************************/

#include <linux/io.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/cnxtnfc.h>
#include "nfc_prv.h"

#define nfc_writel(p,v) writel(v,p)

extern unsigned long ChipRevID; 

inline static int nfc_wait_for_ioseq_ready(void)
{
   NFC_STATUS_WAIT(!(readl(NFC_IO_SEQ_STATUS_REG) & NFC_IO_SEQ_BUSY_MASK));
   return 0;
}

static void nfc_load_ucode(
               const unsigned int   *ucode, 
               const unsigned int   ucode_size)
{
   unsigned int   *nfc_iram_base = (unsigned int*)NFC_IRAM_BASE_REG;
   unsigned int   i     = 0;

   for ( i = 0; i < ucode_size; i++)
   {
      nfc_iram_base[i] = ucode[i];
   }
}

static void nfc_start_cmd_seqr(void)
{
   nfc_writel(NFC_IO_CONFIG_REG, NFC_IO_CFG_PC_STRT);
}

int nfc_default_init(void)
{
   unsigned int   iter;
   int            ret;
   NFC_ERROR_FLAG;

   /**
    * Software workaround for hardware CR 5491 -
    * IO: NAND flash support requires PIO195 to remain configured as an input.
    */
   MAKE_GPIO_INPUT_BANK(6,4);

   /* Needs to take places in PCI initialization code. 
      This will be here till then. */
   /*nfc_writel(PCI_ISAROM_DESC1_REG, 0x2323F050);*/

   /* Load the values to registers. */
   for(iter = 0; iter < NFC_NUM_CONFIG; iter++)
   {
      /* Fill the device configuration register bit fields from DEV_CONFIG 
         structure */
      /* Write the register with filled bit fields. */
      nfc_writel((((LPREG)NFC_DEV_CONFIG0_REG) + iter), 
                  (  (NFC_FLASH_SIZE_2GBITS        << NFC_FLASH_SIZE_SHIFT) |
                     (NFC_PAGE_SIZE_2048BYTES      << NFC_PAGE_SIZE_SHIFT) |
                     (NFC_BUS_WIDTH_16BIT          << NFC_BUS_WIDTH_SHIFT) |
                     (NFC_DEV_TYPE_SLC_LARGE       << NFC_DEV_TYPE_SHIFT) |
                     (NFC_ADDR_CYCLES_5CYCLES      << NFC_ADDR_CYCLES_SHIFT) |
                     (NFC_SPARE_AREA_SIZE_64BYTES  << NFC_SPARE_AREA_SIZE_SHIFT)
                  ));
      /* Write the register with filled bit fields. */
      nfc_writel((((LPREG)NFC_TIMINGS_PARAMS0_REG) + iter), 0xFFFFFFFF);
   }

   /* IO Type interface registers */
   
   /* Write the register with filled bit fields. */
   nfc_writel(NFC_DEV_CONFIG_IO_REG, 0x574);

   iter = readl(NFC_DEV_CONFIG_IO_REG);
   
   /* Write the register with filled bit fields. */
   nfc_writel(NFC_TIMINGS_PARAMS_IO_REG, 0xFFFFFFFF);

   /* Are we booting from NAND flash? */
   if (IS_NANDBOOT)
   {
      /* We have booted from NAND flash - Select Chip 0 */
      nfc_writel(NFC_IO_CS_REG, 0xFE);
   }
   else
   {
      /* We have booted from NOR flash - Select Chip 1 */
     // AIX_3RDPARTY_CODE_CHANGE: chip select is configurable
#ifdef CONFIG_MTD_CNXTNFC_CS
      nfc_writel((NFC_IO_CS_REG), 0xFF & ~(1 << CONFIG_MTD_CNXTNFC_CS) );
#else
      aix: Fehler: bitte CONFIG_MTD_CNXTNFC_CS in kernel config setzen!!)
      nfc_writel((NFC_IO_CS_REG), 0xFD );
#endif
   }

   nfc_writel(NFC_IO_RB_TIMEOUT_REG, 0xFFFFFFFF);

   #if 0
   /* balanid */
   /* AXI-HSX Brigde Timeout setup */
   nfc_writel(0xE8013000, 0x1);
   nfc_writel(0xE8013004, 0x10000000);
   nfc_writel(0xE0450064, 0x2);
   #endif

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   return 0;
}

int nfc_read_id(unsigned int *maker_id, unsigned int *device_id )
{
   int   ret;
   NFC_ERROR_FLAG;

   /* Wait for the flash device to be ready to accept further commands */
   /*NFC_STATUS_WAIT(readl(NFC_RDY_BUSY_STATUS_REG));*/
   /*NFC_STATUS_WAIT(nfc_read_status() & FLASH_STATUS_DEV_OPERATION_MASK);*/

   nfc_writel(NFC_IO_SEQ_GPR0, 0x0);
   nfc_writel(NFC_IO_SEQ_GPR1, 0x0);

   nfc_load_ucode(NFCOpnReadId, NFCOpnReadIdSize);
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   if (NFC_ERROR_OCCURRED)
   {
      *maker_id = *device_id = 0;
      return -EIO;
   }

   *maker_id  = (unsigned int)(readl(NFC_IO_SEQ_GPR0) & 0xFF);
   *device_id = (unsigned int)(readl(NFC_IO_SEQ_GPR1) & 0xFF);
   return 0;
}

#if 0
int nfc_read_status(void)
{
   nfc_load_ucode(NFCOpnReadStatus, NFCOpnReadStatusSize);
   nfc_start_cmd_seqr();

   nfc_wait_for_ioseq_ready();
   
   /* Status is only of 8 bits wide, but returning 32 bits wide value */
   return (readl(NFC_IO_SEQ_GPR0));
}
#endif

int nfc_wait_for_dev_ready(void)
{
   int ret, status;
   NFC_ERROR_FLAG;

   /* Write ONE to GPR0 to get it cleared by the Sequencer. */
   nfc_writel(NFC_IO_SEQ_GPR0, 0x1);

   nfc_load_ucode(NFCOpnWaitForDevRdy, NFCOpnWaitForDevRdySize);
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   /* Wait until the GPR0 is cleared */
   NFC_STATUS_WAIT(!(readl(NFC_IO_SEQ_GPR0)));

   /* Status is only of 8 bits long */
   status = readl(NFC_IO_SEQ_GPR0) & 0xFF;

   if (NFC_ERROR_OCCURRED)
      return -EIO;

   return status;
}

int nfc_dev_reset(unsigned int *status)
{
   int ret;
   NFC_ERROR_FLAG;

   nfc_load_ucode(NFCOpnReset, NFCOpnResetSize);
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   if (status != NULL )
      *status = readl(NFC_IO_SEQ_GPR0);

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   return 0;
}

int nfc_page_read_pio (struct nfc_dev *nfc, unsigned char *buffer, int offset)
{
   unsigned int i, pageSize, spareSize, totalSize;
   int   ret;
   NFC_ERROR_FLAG;

   /* nfc_writel(PCI_INTR_ENABLE_REG, 0x00020000); */
#if 0
repeat:
#endif
   nfc_writel(NFC_FLUSH_REG, (NFC_FLUSH_FIFO_MASK|NFC_FLUSH_BUFFER_MASK));
   if ( 0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }

   /* Load the page number into GPR0. */
   nfc_writel(NFC_IO_SEQ_GPR0, offset);

   /* Configure the DMA_COUNT1 register in units of words (4 bytes) .
      page area is 512 bytes in length for SLC small. */
   pageSize = 0x200; /* 2112/4 = 528 = 0x210 words */
   spareSize = 0x10; 
   totalSize = pageSize + spareSize;

   nfc_writel(NFC_IO_DMA_CNT1_REG, pageSize);
   nfc_writel(NFC_IO_DMA_CNT2_REG, spareSize);

   if (NFC_HAS_WAIT_RB_BUG)
   {
      nfc_load_ucode(NFCOpnReadPage_WA, NFCOpnReadPageSize_WA);
   }
   else
   {
      nfc_load_ucode(NFCOpnReadPage, NFCOpnReadPageSize);
   }
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   #if 0
   /* balanid */
   regval = readl(NFC_IO_PC_REG);
   while ((NFCOpnReadPageSize-1) != regval ){
      nfc_writel(NFC_IO_CONFIG_REG, 0x02);
      while(i < 1000000) i++;
      nfc_start_cmd_seqr();
      nfc_wait_for_ioseq_ready();
      regval = readl(NFC_IO_PC_REG);
   }
   #endif
   
   /* Data is available in FIFO. Read it into the array. */
   /* IMPORTANT: DO NOT READ EXTRA WORDS FROM FIFO. 
      IT WILL HANG UNTIL THE DATA IS AVAILABLE. */
    for (i = 0; i < totalSize; i++)
   {
      ((unsigned int*)buffer)[i]= readl(NFC_IO_SEQ_FIFO_REG);
      #if 0
      /* balanid */
      if ((readl(0xE0450070) & 0x2) != 0)
      {
         nfc_writel(0xE8013000, 0x0);
         nfc_writel(0xE8013004, 0xFFFFFFFF);
         goto repeat;
      }
      #endif
   }

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   return totalSize*4;
}

int nfc_page_read_dma (struct nfc_dev *nfc, unsigned char *buffer, int offset)
{
   unsigned int   pagesize, sparesize, totalsize;
   dma_addr_t     udata;
   int            ret;
   NFC_ERROR_FLAG;

   /* nfc_writel(PCI_INTR_ENABLE_REG, 0x00020000); */

   nfc_writel(NFC_FLUSH_REG, (NFC_FLUSH_FIFO_MASK|NFC_FLUSH_BUFFER_MASK));
   if ( 0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }

   /* Load the page number into GPR0. */
   nfc_writel(NFC_IO_SEQ_GPR0, offset);

   /* Configure the DMA_COUNT1 register in units of words (4 bytes) .
      page area is 512 bytes in length for SLC small. */

   /* 2112/4 = 528 = 0x210 words */
   pagesize    = 0x200; /* 0x800 = 2048 Bytes */
   sparesize   = 0x10;  /* 0x40  = 64 Bytes */
   totalsize   = pagesize + sparesize;

   nfc_writel(NFC_IO_DMA_CNT1_REG, pagesize);
   nfc_writel(NFC_IO_DMA_CNT2_REG, sparesize);

   if (NFC_DMA_ALLOC_COHERENT == NO)
   {
      nfc->dmaaddr = 
         dma_map_single(NULL, nfc->page_buf, totalsize*4, DMA_FROM_DEVICE);
   }

   nfc_dma_read_setup(nfc->dmaaddr, pagesize*4, 
                        DMA_CTRL_REQ_PCI1, DMA_CHANNEL6);
   nfc_dma_start(DMA_CHANNEL6);

   udata = (nfc->dmaaddr + (pagesize*4));

   nfc_dma_read_setup(udata, sparesize*4, 
                        DMA_CTRL_REQ_PCI2, DMA_CHANNEL7);
   nfc_dma_start(DMA_CHANNEL7);

   nfc_load_ucode(NFCOpnReadPage, NFCOpnReadPageSize);
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   ret = nfc_dma_wait2complete(DMA_CHANNEL6);
   NFC_ERROR_CHECK(ret);
   nfc_dma_clear(DMA_CHANNEL6);

   ret = nfc_dma_wait2complete(DMA_CHANNEL7);
   NFC_ERROR_CHECK(ret);
   nfc_dma_clear(DMA_CHANNEL7);

   if (NFC_DMA_ALLOC_COHERENT == NO)
   {
      dma_unmap_single(NULL, nfc->dmaaddr, totalsize*4, DMA_FROM_DEVICE);
   }

   if (NFC_ERROR_OCCURRED)
      return -EIO;

   return (totalsize*4);
}

int nfc_page_read (struct nfc_dev *nfc, unsigned char *buffer, int offset)
{
   if (NFC_DMA_READ_ENABLE)
   {
      return nfc_page_read_dma(nfc,buffer,offset);
   }
   else
   {
      return nfc_page_read_pio(nfc,buffer,offset);
   }
}

int nfc_page_write_pio (struct nfc_dev *nfc, const unsigned char *buffer, int offset)
{
   unsigned int i, pagesize, sparesize, totalsize, status;
   int ret;
   NFC_ERROR_FLAG;

   /* nfc_writel(PCI_INTR_ENABLE_REG, 0x00020000); */

   nfc_writel(NFC_FLUSH_REG, (NFC_FLUSH_FIFO_MASK|NFC_FLUSH_BUFFER_MASK));

   if ( 0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }

   /* Load the page number into GPR0. */
   nfc_writel(NFC_IO_SEQ_GPR0, offset);

   /* Configure the DMA_COUNT1 register in units of words (4 bytes) .
      page area is 512 bytes in length for SLC small. */

   pagesize    = 0x200; /* 0x800 = 2048 Bytes */
   sparesize   = 0x0;   /* ECC will be written in the last word */
   totalsize   = pagesize + sparesize;

   nfc_writel(NFC_IO_DMA_CNT1_REG, pagesize);
   nfc_writel(NFC_IO_DMA_CNT2_REG, sparesize);

   /* Write buffer to FIFO. */
   #if 0
   nfc_dma_write_setup((unsigned int)buffer, totalsize*4, 0, 0);
   nfc_dma_start(0);
   nfc_dma_wait2complete(0);
   #else
   for (i = 0; i < totalsize; i++)
      nfc_writel(NFC_IO_SEQ_FIFO_REG, ((unsigned int*)buffer)[i]);
   #endif

   nfc_dev_unlock();

   if (NFC_HAS_WAIT_RB_BUG)
   {
      nfc_load_ucode(NFCOpnPageWrite_WA, NFCOpnPageWriteSize_WA);
   }
   else
   {
      nfc_load_ucode(NFCOpnPageWrite, NFCOpnPageWriteSize);
   }
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   #if 0
   regval = readl(NFC_IO_PC_REG);
   while ((NFCOpnPageWriteSize-1) != regval ){
       nfc_writel(NFC_IO_CONFIG_REG, 0x02);
       udelay(2000);
       nfc_start_cmd_seqr();
       nfc_wait_for_ioseq_ready();
       regval = readl(NFC_IO_PC_REG);
   }
   #endif

   nfc_dev_lock();

   status = readl(NFC_IO_SEQ_GPR0);

   if (NFC_ERROR_OCCURRED)
      return -EIO;

   return (status & FLASH_STATUS_PROGRAM_ERASE_MASK);
}

int nfc_page_write_dma (struct nfc_dev *nfc, const unsigned char *buffer, int offset)
{
   unsigned int pagesize, sparesize, totalsize, status;
   int ret;
   NFC_ERROR_FLAG;

   /* nfc_writel(PCI_INTR_ENABLE_REG, 0x00020000); */

   nfc_writel(NFC_FLUSH_REG, (NFC_FLUSH_FIFO_MASK|NFC_FLUSH_BUFFER_MASK));
   if ( 0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }

   /* Load the page number into GPR0. */
   nfc_writel(NFC_IO_SEQ_GPR0, offset);

   /* Configure the DMA_COUNT1 register in units of words (4 bytes) .
      page area is 512 bytes in length for SLC small. */

   pagesize    = 0x200; /* 0x800 = 2048 Bytes */
   sparesize   = 0x0;   /* ECC will be written in the last word */
   totalsize   = pagesize + sparesize;

   nfc_writel(NFC_IO_DMA_CNT1_REG, pagesize);
   nfc_writel(NFC_IO_DMA_CNT2_REG, sparesize);

   if (NFC_DMA_ALLOC_COHERENT == NO)
   {
      nfc->dmaaddr = 
         dma_map_single(NULL, nfc->page_buf, totalsize*4, DMA_TO_DEVICE);
   }

   /* Write buffer to FIFO. */
   nfc_dma_write_setup(nfc->dmaaddr, pagesize*4, 
                        DMA_CTRL_REQ_CONSTANT, DMA_CHANNEL6);
   nfc_dma_start(DMA_CHANNEL6);

   ret = nfc_dma_wait2complete(DMA_CHANNEL6);
   NFC_ERROR_CHECK(ret);

   nfc_dma_clear(DMA_CHANNEL6);

   if (NFC_DMA_ALLOC_COHERENT == NO)
   {
      dma_unmap_single(NULL, nfc->dmaaddr, totalsize, DMA_TO_DEVICE);
   }

   nfc_dev_unlock();

   nfc_load_ucode(NFCOpnPageWrite, NFCOpnPageWriteSize);
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   nfc_dev_lock();

   status = readl(NFC_IO_SEQ_GPR0);

   if (NFC_ERROR_OCCURRED)
      return -EIO;

   return (status & FLASH_STATUS_PROGRAM_ERASE_MASK);
}

int nfc_page_write (struct nfc_dev *nfc, unsigned char *buffer, int offset)
{
   if (NFC_DMA_WRITE_ENABLE)
   {
      return nfc_page_write_dma(nfc,buffer,offset);
   }
   else
   {
      return nfc_page_write_pio(nfc,buffer,offset);
   }
}

void nfc_dev_unlock(void)
{
   /* unlock the registers */
   nfc_writel(NFC_UNLOCK_DES0_REG,0xf00200a8);
   nfc_writel(NFC_UNLOCK_DES0_REG,0x0300E028);
   nfc_writel(NFC_UNLOCK_DES1_REG,0xf0120670);
   nfc_writel(NFC_UNLOCK_DES1_REG,0x03f0e1e9);
   nfc_writel(NFC_UNLOCK_DES2_REG,0xe93202e8);
   nfc_writel(NFC_UNLOCK_DES2_REG,0xd30fe100);
   nfc_writel(NFC_UNLOCK_DES3_REG,0xa0d202a9);
   nfc_writel(NFC_UNLOCK_DES3_REG,0xb3e3e136);
   nfc_writel(NFC_IO_UNLOCK_REG,0x11E313b0);
   nfc_writel(NFC_IO_UNLOCK_REG,0x070993a0);

   /* write start addr & end addr -----??? They should be other way */
   nfc_writel(NFC_LOCK_STRT_ADDR_DES0_REG,0xffffffff);
   nfc_writel(NFC_LOCK_END_ADDR_DES0_REG,0xf0000000);
   nfc_writel(NFC_LOCK_STRT_ADDR_DES1_REG,0xffffffff);
   nfc_writel(NFC_LOCK_END_ADDR_DES1_REG,0xf0000000);
   nfc_writel(NFC_LOCK_STRT_ADDR_DES2_REG,0xffffffff);
   nfc_writel(NFC_LOCK_END_ADDR_DES2_REG,0xf0000000);
   nfc_writel(NFC_LOCK_STRT_ADDR_DES3_REG,0xffffffff);
   nfc_writel(NFC_LOCK_END_ADDR_DES3_REG,0xf0000000);
   nfc_writel(NFC_IO_LOCK_STRT_ADDR_REG,0xffffffff);
   nfc_writel(NFC_IO_LOCK_END_ADDR_REG, 0xf0000000);
}

void nfc_dev_lock(void)
{
   /* unlock the registers */
   nfc_writel(NFC_UNLOCK_DES0_REG,0x0);
   nfc_writel(NFC_UNLOCK_DES1_REG,0x0);
   nfc_writel(NFC_UNLOCK_DES2_REG,0x0);
   nfc_writel(NFC_UNLOCK_DES3_REG,0x0);
   nfc_writel(NFC_IO_UNLOCK_REG,0x0);

   /* write start addr & end addr -----??? They should be other way */
   nfc_writel(NFC_LOCK_STRT_ADDR_DES0_REG,0x0);
   nfc_writel(NFC_LOCK_END_ADDR_DES0_REG,0x0);
   nfc_writel(NFC_LOCK_STRT_ADDR_DES1_REG,0x0);
   nfc_writel(NFC_LOCK_END_ADDR_DES1_REG,0x0);
   nfc_writel(NFC_LOCK_STRT_ADDR_DES2_REG,0x0);
   nfc_writel(NFC_LOCK_END_ADDR_DES2_REG,0x0);
   nfc_writel(NFC_LOCK_STRT_ADDR_DES3_REG,0x0);
   nfc_writel(NFC_LOCK_END_ADDR_DES3_REG,0x0);
   nfc_writel(NFC_IO_LOCK_STRT_ADDR_REG,0x0);
   nfc_writel(NFC_IO_LOCK_END_ADDR_REG, 0x0);
}

int nfc_block_erase( struct nfc_dev *nfc, int  offset)
{
   int ret;
   NFC_ERROR_FLAG;

   unsigned int status;

   nfc_writel(NFC_IO_SEQ_GPR0, offset);
   
   if (NFC_HAS_WAIT_RB_BUG)
   {
      nfc_load_ucode(NFCOpnBlockErase_WA, NFCOpnBlockEraseSize_WA);
   }
   else
   {
      nfc_load_ucode(NFCOpnBlockErase, NFCOpnBlockEraseSize);
   }
   nfc_dev_unlock();
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   nfc_dev_lock();
   status = readl(NFC_IO_SEQ_GPR0);

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   if (  (status & FLASH_STATUS_DEV_OPERATION_MASK) && 
        !(status & FLASH_STATUS_PROGRAM_ERASE_MASK))
      return 0;

   return 1;
}

int nfc_spare_read( struct nfc_dev *nfc, unsigned char *buffer, int offset)
{
   unsigned int i, spareSize;
   int ret;
   NFC_ERROR_FLAG;

   /* Load the page number into GPR0. */
#if 0
repeat:
#endif
   nfc_writel(NFC_FLUSH_REG, (NFC_FLUSH_FIFO_MASK|NFC_FLUSH_BUFFER_MASK));

   if ( 0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }
   
   nfc_writel(NFC_IO_SEQ_GPR0, offset);

   /* nfc_writel(PCI_INTR_ENABLE_REG, 0x00020000); */

   /* Configure the DMA_COUNT1 register in units of words (4 bytes) .
      page area is 512 bytes in length for SLC small. */
   spareSize = 0x10; 

   nfc_writel(NFC_IO_DMA_CNT1_REG, spareSize);
   nfc_writel(NFC_IO_DMA_CNT2_REG, 0x0);

   if(NFC_HAS_WAIT_RB_BUG)
   {
      nfc_load_ucode(NFCOpnReadSpare_WA, NFCOpnReadSpareSize_WA);
   }
   else
   {
      nfc_load_ucode(NFCOpnReadSpare, NFCOpnReadSpareSize);
   }
   nfc_start_cmd_seqr();
   
   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   #if 0
   /* balanid */
   regval = readl(NFC_IO_PC_REG);
   while ((NFCOpnReadSpareSize-1) != regval ){
      nfc_writel(NFC_IO_CONFIG_REG, 0x02);
      while(i < 1000000) i++;
      nfc_start_cmd_seqr();
      nfc_wait_for_ioseq_ready();
      regval = readl(NFC_IO_PC_REG);
   }
   #endif
   
   /* Data is available in FIFO. Read it into the array. */
   /* IMPORTANT: DO NOT READ EXTRA WORDS FROM FIFO. 
      IT WILL HANG UNTIL THE DATA IS AVAILABLE. */
   for (i = 0; i < spareSize; i++)
   {
      ((unsigned int*)buffer)[i]= readl(NFC_IO_SEQ_FIFO_REG);
      #if 0
      /* balanid */
      if ((readl(0xE0450070) & 0x2) != 0)
      {
         nfc_writel(0xE8013000, 0x0);
         nfc_writel(0xE8013004, 0xFFFFFFFF);
         goto repeat;
      }
      #endif
   }

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   return 0;
}

int nfc_spare_write(struct nfc_dev *nfc, const unsigned char *buffer, int offset)
{
   unsigned int i, spareSize, status;
   int ret;
   NFC_ERROR_FLAG;

   nfc_writel(NFC_FLUSH_REG, (NFC_FLUSH_FIFO_MASK|NFC_FLUSH_BUFFER_MASK));

   if ( 0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }

   /* Load the page number into GPR0. */
   nfc_writel(NFC_IO_SEQ_GPR0, offset);

   /* Configure the DMA_COUNT1 register in units of words (4 bytes) .
      page area is 512 bytes in length for SLC small. */
   spareSize = 0x10; /* ECC will be written in the last 4 words */
 
   nfc_writel(NFC_IO_DMA_CNT1_REG, spareSize);
   nfc_writel(NFC_IO_DMA_CNT2_REG, 0x0);

   /* Write buffer to FIFO. */
   #if 0
   internal_nfc_dma_write_setup((u_int32)buffer, totalSize*4, 0, 0);
   internal_nfc_dma_start(0);
   internal_nfc_dma_wait2complete(0);
   #else
   for (i = 0; i < spareSize; i++)
      nfc_writel(NFC_IO_SEQ_FIFO_REG, ((unsigned int*)buffer)[i]);
   #endif

   if (NFC_HAS_WAIT_RB_BUG)
   {
      nfc_load_ucode(NFCOpnSpareWrite_WA, NFCOpnSpareWriteSize_WA);
   }
   else
   {
      nfc_load_ucode(NFCOpnSpareWrite, NFCOpnSpareWriteSize);
   }

   nfc_dev_unlock();
   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   #if 0
   regval = readl(NFC_IO_PC_REG);
   while ((NFCOpnSpareWriteSize-1) != regval ){
      printk("nfc_spare_write 7: regval = %d\n", regval);  
      nfc_writel(NFC_IO_CONFIG_REG, 0x02);
      udelay(2000);
      nfc_start_cmd_seqr();
      nfc_wait_for_ioseq_ready();
      regval = readl(NFC_IO_PC_REG);
      nfc_dev_lock();
      return -EIO;
   }
   #endif
   nfc_dev_lock();

   status = readl(NFC_IO_SEQ_GPR0);

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   return (status & FLASH_STATUS_PROGRAM_ERASE_MASK);
}

int nfc_check_isbad( struct nfc_dev *nfc, unsigned int pgaddr )
{
   int ret;
   NFC_ERROR_FLAG;

   /* nfc_writel(PCI_INTR_ENABLE_REG, 0x00020000); */

   if (0 == readl(NFC_RDY_BUSY_STATUS_REG))
   {
      ret = nfc_dev_reset(NULL);
      NFC_ERROR_CHECK(ret);
   }
   /* Load the page number into GPR0. */
   nfc_writel(NFC_IO_SEQ_GPR0, pgaddr>>nfc->page_shift);

   if (NFC_HAS_WAIT_RB_BUG)
   {
      nfc_load_ucode(NFCOpnReadBadBlkMarker_WA, NFCOpnReadBadBlkMarkerSize_WA);
   }
   else
   {
      nfc_load_ucode(NFCOpnReadBadBlkMarker, NFCOpnReadBadBlkMarkerSize);
   }

   nfc_start_cmd_seqr();

   ret = nfc_wait_for_ioseq_ready();
   NFC_ERROR_CHECK(ret);

   #if 0
   regval = readl(NFC_IO_PC_REG);
   while ((NFCOpnReadBadBlkMarkerSize-1) != regval ){
       nfc_writel(NFC_IO_CONFIG_REG, 0x02);
       udelay(2000);
       nfc_start_cmd_seqr();
       nfc_wait_for_ioseq_ready();
       regval = readl(NFC_IO_PC_REG);
   }
   #endif

   if(NFC_ERROR_OCCURRED)
      return -EIO;

   if ( 0xFF != (readl(NFC_IO_SEQ_GPR1) & 0xFF) 
      || 0xFF != (readl(NFC_IO_SEQ_GPR2) & 0xFF))
      return 1;

   return 0;
}

void nfc_dma_write_setup(
               unsigned int   src_addr,
               unsigned int   byte_count,
               unsigned int   request_sel,
               unsigned int   channel)
{
   unsigned ctrl = 0;

   /* Fill in the Channel registers with SRC and DEST address
      and recovery value */
   nfc_writel(DMA_SRC_BASE_REG_CH(channel), src_addr);
   nfc_writel(DMA_DEST_BASE_ADDR_REG_CH(channel), NFC_IO_SEQ_FIFO_REG);
   nfc_writel(DMA_RECOVERY_REG_CH(channel), 0x2A); /* ?? */

   /*SRC and DEST Run/Size */
   nfc_writel(DMA_SRC_RUN_BUF_SIZE_REG_CH(channel), byte_count);
   nfc_writel(DMA_DEST_RUN_BUF_SIZE_REG_CH(channel), byte_count);

   /*SRC and DEST Num strides */
   nfc_writel(DMA_SRC_NUM_STRIDES_REG_CH(channel), 0x1);
   nfc_writel(DMA_DEST_NUM_STRIDE_BUF_REG_CH(channel), 0x1);


   /*SRC and DEST stride interval  */
   nfc_writel(DMA_SRC_STRIDE_INT_REG_CH(channel), 0x0);
   nfc_writel(DMA_DEST_STRIDE_INT_REG_CH(channel), 0x0);


   /* Write to values to CONTROL register.
    *
    * [23:16]  Peripheral DMA request select (0x0)
    * [9]      DO NOT increment destination address (0)
    * [8]      Increment source address (1)
    * [7:6]    Dest xfer size (10 = word)
    * [5:4]    Source xfer size (10 = word)
    * [3]      Buffer xfer mode (1 = stop at end of buffer)
    *
    * *** DONOT START THE DMA YET. ***
    */
   ctrl = ( request_sel |
            DMA_CTRL_INC_DEST_ADDR_INC_NO |
            DMA_CTRL_INC_SRC_ADDR_INC_YES |
            DMA_CTRL_SRC_XFER_SIZE_WORD |
            DMA_CTRL_DEST_XFER_SIZE_WORD |
            DMA_CTRL_XFER_MODE_STOP);
   nfc_writel(DMA_CTRL_REG_CH(channel), ctrl);
}

void nfc_dma_read_setup(
               unsigned int   dest_addr,
               unsigned int   byte_count,
               unsigned int   request_sel,
               unsigned int   channel)
{
   unsigned ctrl = 0;

   /* Fill in the Channel registers with SRC and DEST address
    * and recovery value */
   nfc_writel(DMA_SRC_BASE_REG_CH(channel), NFC_IO_SEQ_FIFO_REG);
   nfc_writel(DMA_DEST_BASE_ADDR_REG_CH(channel), dest_addr);
   nfc_writel(DMA_RECOVERY_REG_CH(channel), 0x2A); /* ?? */

   /*SRC and DEST Run/Size */
   nfc_writel(DMA_SRC_RUN_BUF_SIZE_REG_CH(channel), byte_count);
   nfc_writel(DMA_DEST_RUN_BUF_SIZE_REG_CH(channel), byte_count);

   /*SRC and DEST Num strides */
   nfc_writel(DMA_SRC_NUM_STRIDES_REG_CH(channel), 0x1);
   nfc_writel(DMA_DEST_NUM_STRIDE_BUF_REG_CH(channel), 0x1);

   /*SRC and DEST stride interval  */
   nfc_writel(DMA_SRC_STRIDE_INT_REG_CH(channel), 0x0);
   nfc_writel(DMA_DEST_STRIDE_INT_REG_CH(channel), 0x0);

   /* Write to values to CONTROL register.
    *
    * [23:16]  Peripheral DMA request select (0x31 = 49)
    * [9]      Increment Destination address (1)
    * [8]      DO NOT increment source address (0)
    * [7:6]    Dest xfer size (10 = word)
    * [5:4]    Source xfer size (10 = word)
    * [3]      Buffer xfer mode (1 = stop at end of buffer)
    *
    * *** DONOT START THE DMA YET. ***
    */
   ctrl = ( request_sel |
            DMA_CTRL_INC_DEST_ADDR_INC_YES |
            DMA_CTRL_INC_SRC_ADDR_INC_NO |
            DMA_CTRL_DEST_XFER_SIZE_WORD |
            DMA_CTRL_SRC_XFER_SIZE_WORD |
            DMA_CTRL_XFER_MODE_STOP);
   nfc_writel(DMA_CTRL_REG_CH(channel),ctrl);
}

void nfc_dma_start(unsigned int channel)
{
   unsigned int ctrl = 0;
   /* Everything is setup. 
    * Just write Bit 0 of Channel control register to HIGH*/
   ctrl = (readl(DMA_CTRL_REG_CH(channel)) | DMA_CTRL_ENABLE);
   nfc_writel(DMA_CTRL_REG_CH(channel),ctrl);
}

int nfc_dma_wait2complete(unsigned int channel)
{
   NFC_STATUS_WAIT( (readl(DMA_STATUS_REG_CH(channel))) & 
                     (  DMA_STATUS_DMA_SRC_BUFF_INT | 
                        DMA_STATUS_DMA_DEST_BUFF_INT));
   return 0;
}

void nfc_dma_clear(unsigned int channel)
{
   unsigned ctrl ;

   ctrl = readl(DMA_CTRL_REG_CH(channel));
   ctrl = ctrl & (~DMA_CTRL_ENABLE);
   nfc_writel(DMA_CTRL_REG_CH(channel),ctrl);
}

/****************************************************************************
 * Modifications:
 * $Log:
 * $
 *
 ****************************************************************************/
