/*
 * cnxt_otg_transiever.c - Conexant USB OTG transiever driver
 *
 * Copyright (C) 2007 Conexant Systems Inc
 * Copyright (C) 2007 P S Senthil Kumar
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "cnxt_otg_transceiver.h"

/* OTG Transiever Interface Global variables*/
static struct otg_transceiver *xceiv0,*xceiv1;
int trans;

void dcd_power(u8 controller_id,unsigned mA) 
{

	/* Device Controller port doesn't support port power control */
}

/*-------------------------------------------------------------------------*/

static const char *state_string(enum usb_otg_state state)
{
	switch (state) {
	case OTG_STATE_A_IDLE:		return "a_idle";
	case OTG_STATE_A_WAIT_VRISE:	return "a_wait_vrise";
	case OTG_STATE_A_WAIT_BCON:	return "a_wait_bcon";
	case OTG_STATE_A_HOST:		return "a_host";
	case OTG_STATE_A_SUSPEND:	return "a_suspend";
	case OTG_STATE_A_PERIPHERAL:	return "a_peripheral";
	case OTG_STATE_A_WAIT_VFALL:	return "a_wait_vfall";
	case OTG_STATE_A_VBUS_ERR:	return "a_vbus_err";
	case OTG_STATE_B_IDLE:		return "b_idle";
	case OTG_STATE_B_SRP_INIT:	return "b_srp_init";
	case OTG_STATE_B_PERIPHERAL:	return "b_peripheral";
	case OTG_STATE_B_WAIT_ACON:	return "b_wait_acon";
	case OTG_STATE_B_HOST:		return "b_host";
	default:			return "UNDEFINED";
	}
}

static inline const char *state_name(struct cnxt_otg_trans *cnxt_otg)
{
	return state_string(cnxt_otg->otg.state);
}

#ifdef	VERBOSE
#define	dev_vdbg			dev_dbg
#else
#define	dev_vdbg(dev, fmt, arg...)	do{}while(0)
#endif




/**
 * otg_get_transceiver - find the OTG transceiver driver
 *
 * Returns the transceiver driver, after getting a refcount to it; or
 * null if there is no such transceiver.  The caller is responsible for
 * releasing that count.
 */
struct otg_transceiver *otg_get_transceiver(void)
{

  if(trans == OTG_TRANSIEVER_ID_0)
  {
	if (xceiv0)
		get_device(xceiv0->dev);
	return xceiv0;

  }else if(trans == OTG_TRANSIEVER_ID_1)
  {
	if (xceiv1)
		get_device(xceiv1->dev);
	return xceiv1;
  }
  else
  {
	printk(" otg_get_transceiver failed ...\n");
	return NULL;

  }

}


EXPORT_SYMBOL(otg_get_transceiver);

int otg_set_transceiver(struct otg_transceiver *x)
{
	if ( (xceiv0 && x) || ( (xceiv1 && x) ))
		return -EBUSY;
	
	if(trans == OTG_TRANSIEVER_ID_0)
	{
		xceiv0 = x;
	}
	else if(trans == OTG_TRANSIEVER_ID_1)
	{
		xceiv1 = x;
	}
	return 0;
}
EXPORT_SYMBOL(otg_set_transceiver);





static void power_down(struct cnxt_otg_trans *cnxt_otg)
{
	cnxt_otg->otg.state = OTG_STATE_UNDEFINED;

	dcd_power(cnxt_otg->otg_controller_id,0); /* Set Device Controller to suspend mode */
}


#define	NO_HOST_SUSPEND

static int host_suspend(struct cnxt_otg_trans *cnxt_otg)
{
#ifdef	NO_HOST_SUSPEND
	return 0;
#else
	struct device	*dev;

	if (!cnxt_otg->otg.host)
		return -ENODEV;

	/* Currently ASSUMES only the OTG port matters;
	 * other ports could be active...
	 */
	dev = cnxt_otg->otg.host->controller;
	return dev->driver->suspend(dev, 3, 0);
#endif
}

static int host_resume(struct cnxt_otg_trans *cnxt_otg)
{
#ifdef	NO_HOST_SUSPEND
	return 0;
#else
	struct device	*dev;

	if (!cnxt_otg->otg.host)
		return -ENODEV;

	dev = cnxt_otg->otg.host->controller;
	return dev->driver->resume(dev, 0);
#endif
}

static int gadget_suspend(struct cnxt_otg_trans *cnxt_otg)
{
	cnxt_otg->otg.gadget->b_hnp_enable = 0;
	cnxt_otg->otg.gadget->a_hnp_support = 0;
	cnxt_otg->otg.gadget->a_alt_hnp_support = 0;
	return usb_gadget_vbus_disconnect(cnxt_otg->otg.gadget);
}

/*-------------------------------------------------------------------------*/


/* 
 * NOTE: guaranteeing certain response times might mean we shouldn't
 * share keventd's work queue; a realtime task might be safest.
 */
void
cnxt_otg_defer_work(struct cnxt_otg_trans *cnxt_otg, int work)
{
	int status;

	if (cnxt_otg && !test_and_set_bit(work, &cnxt_otg->todo)) {
		
		status = schedule_work(&cnxt_otg->work);
		if (!status && !cnxt_otg->working)
			dev_vdbg("work item %d may be lost\n", work);
	}
}

/* called from irq handlers */
static void a_idle(struct cnxt_otg_trans *cnxt_otg, const char *tag)
{
	if (cnxt_otg->otg.state == OTG_STATE_A_IDLE)
		return;

	cnxt_otg->otg.default_a = 1;
	if (cnxt_otg->otg.host) {
		cnxt_otg->otg.host->is_b_host = 0;
		host_suspend(cnxt_otg);
	}
	if (cnxt_otg->otg.gadget) {
		cnxt_otg->otg.gadget->is_a_peripheral = 1;
		gadget_suspend(cnxt_otg);
	}
	cnxt_otg->otg.state = OTG_STATE_A_IDLE;
		pr_debug("  --> %s/%s\n", state_name(cnxt_otg), tag);
}

/* called from irq handlers */
static void b_idle(struct cnxt_otg_trans *cnxt_otg, const char *tag)
{
	if (cnxt_otg->otg.state == OTG_STATE_B_IDLE)
		return;

	cnxt_otg->otg.default_a = 0;
	if (cnxt_otg->otg.host) {
		cnxt_otg->otg.host->is_b_host = 1;
		host_suspend(cnxt_otg);
	}
	if (cnxt_otg->otg.gadget) {
		cnxt_otg->otg.gadget->is_a_peripheral = 0;
		gadget_suspend(cnxt_otg);
	}
	cnxt_otg->otg.state = OTG_STATE_B_IDLE;
	pr_debug("  --> %s/%s\n", state_name(cnxt_otg), tag);
}



/*-------------------------------------------------------------------------*/

#ifdef	CONFIG_USB_OTG

 



static irqreturn_t cnxt_otg_irq(int irq, void *_cnxt_otg, struct pt_regs *regs)
{
	int		ret = IRQ_NONE;
	struct cnxt_otg_trans	*cnxt_otg = _cnxt_otg;
	u32 cmd;


	cmd = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));

	/* Detect ID interrupt  and enable the correspoding otg controller */
	
	if( cmd & OTG_ID_IS)
	{
		

		if( cmd & OTD_ID)
		{
			
			int	change = 0;

			cmd |= OTG_ID_IS;
		    (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */


			if (cnxt_otg->otg.host) {
				switch (cnxt_otg->otg.state) {
				case OTG_STATE_B_WAIT_ACON:
					cnxt_otg->otg.state = OTG_STATE_B_HOST;
					pr_debug("  --> b_host\n");
					change = 1;
					break;
				case OTG_STATE_A_WAIT_BCON:
					cnxt_otg->otg.state = OTG_STATE_A_HOST;
					pr_debug("  --> a_host\n");
					break;
				case OTG_STATE_A_PERIPHERAL:
					cnxt_otg->otg.state = OTG_STATE_A_WAIT_BCON;
					pr_debug("  --> a_wait_bcon\n");
					break;
				default:
					break;
				}
				cnxt_otg_defer_work(cnxt_otg, WORK_HOST_RESUME);
				ret = IRQ_HANDLED;

			if (kick)
				usb_bus_start_enum(cnxt_otg->otg.host,
						cnxt_otg->otg.host->otg_port);
			}

		}
		else
		{
			/* SENTHIL : TODO --- DEVICE MODE CODE  */
			printk("  device state \n");
		}

		cmd |= OTG_ID_IS;
		(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */



	}
	elseif(cmd & OTG_DPS) /* SRP DETECT from B-device */ 
	{


		cnxt_otg_defer_work(cnxt_otg, WORK_UPDATE_OTG);
		switch (cnxt_otg->otg.state) {
		case OTG_STATE_A_IDLE:
			if (!cnxt_otg->otg.host)
				break;
			cnxt_otg_defer_work(cnxt_otg, WORK_HOST_RESUME);
		
			break;
		default:
			break;
		}

		
		ret = IRQ_HANDLED;
	
	}
	else if ( cmd  & (  OTG_BSE_IS | OTG_BSV_IS | OTG_ASV_IS | OTG_AVV_IS ) )
	{
		cnxt_otg_defer_work(cnxt_otg, WORK_UPDATE_OTG);
		
	}

	return ret;




}



static struct platform_device *otg_dev;



static int otg_probe(struct platform_device *dev)
{
	// struct omap_usb_config *config = dev->platform_data;

	otg_dev = dev;
	return 0;
}

static int otg_remove(struct platform_device *dev)
{
	otg_dev = 0;
	return 0;
}

struct platform_driver cnxt_otg_driver = {
	.probe		= otg_probe,
	.remove		= otg_remove,
	.driver		= {
		.owner	= THIS_MODULE,
		.name	= "cnxt_otg",
	},
};

static int otg_bind(struct cnxt_otg_trans *cnxt_otg)
{
	int	status;

	if (otg_dev)
		return -EBUSY;

	status = platform_driver_register(&cnxt_otg_driver);
	if (status < 0)
		return status;



	if (otg_dev)
	{
		status = request_irq(cnxt_otg->irq, cnxt_otg_irq,
				IRQF_SHARED, DRIVER_NAME, cnxt_otg);

	}
	else
		status = -ENODEV;

	if (status < 0)
		platform_driver_unregister(&cnxt_otg_driver);
	return status;
}

static void otg_unbind(struct cnxt_otg_irq *cnxt_otg)
{
	if (!otg_dev)
		return;
	
}

#else

/* OTG controller isn't clocked */

#endif	/* CONFIG_USB_OTG */

/*-------------------------------------------------------------------------*/


/* SENTHIL : TODO DEVICE MODE CODE - InComplete */

static void b_peripheral(struct cnxt_otg_trans *cnxt_otg)
{
	
	usb_gadget_vbus_connect(cnxt_otg->otg.gadget);

#ifdef	CONFIG_USB_OTG
	otg_update_cnxt(cnxt_otg);
#else
	
	/* UDC driver just set OTG_BSESSVLD register - TODO*/
	
	cnxt_otg->otg.state = OTG_STATE_B_PERIPHERAL;
	pr_debug("  --> b_peripheral\n");
	
#endif
}

static void cnxt_update_otg(struct cnxt_otg_trans *cnxt_otg)
{
	u32			cnxt_stat, cnxt_bstat,cmd;
	enum usb_otg_state	state = cnxt_otg->otg.state;

	cmd = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));


    if( cmd & OTG_BSE_IS)
	{
		cnxt_otg->state_internal = OTG_STATE_INTERNAL_B_SESSION_END ;
		cmd |= OTG_BSE_IS;
		(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */

	}else if( cmd  & OTG_BSV_IS)
	{
		cnxt_otg->state_internal = OTG_STATE_INTERNAL_B_SESSION_VALID ;
		cmd |= OTG_BSV_IS;
		(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */

	}else if (cmd  & OTG_ASV_IS)
	{
		cnxt_otg->state_internal = OTG_STATE_INTERNAL_A_SESSION_VALID ;
		cmd |= OTG_ASV_IS;
		(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */

	}else if (cmd  & OTG_AVV_IS)
	{
		cnxt_otg->state_internal = OTG_STATE_INTERNAL_A_VBUS_END ;
		cmd |= OTG_AVV_IS;
		(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */

	}else if (cmd & OTG_ID) {
		if (cnxt_otg->otg.default_a) {
			switch (state) {
			case OTG_STATE_B_IDLE:
				a_idle(cnxt_otg, "idle");
				/* FALLTHROUGH */
			case OTG_STATE_A_IDLE:
				/* FALLTHROUGH */
			case OTG_STATE_A_WAIT_VRISE:
				/* we skip over OTG_STATE_A_WAIT_BCON, since
				 * the HC will transition to A_HOST (or
				 * A_SUSPEND!) without our noticing except
				 * when HNP is used.
				 */
				if (cmd & OTG_AVV)
					cnxt_otg->otg.state = OTG_STATE_A_HOST;
				break;
			case OTG_STATE_A_WAIT_VFALL:
				if (!(cmd & OTG_ASV))
					a_idle(cnxt_otg, "vfell");
				break;
			default:
				if (!(cmd  & OTG_AVV))
					cnxt_otg->otg.state = OTG_STATE_A_VBUS_ERR;
				break;
			}
			
		} else {
			switch (state) {
			case OTG_STATE_B_PERIPHERAL:
			case OTG_STATE_B_HOST:
			case OTG_STATE_B_WAIT_ACON:
				usb_gadget_vbus_disconnect(cnxt_otg->otg.gadget);
				break;
			default:
				break;
			}
			if (state != OTG_STATE_A_IDLE)
				a_idle(cnxt_otg, "id");
			if (cnxt_otg->otg.host && state == OTG_STATE_A_IDLE)
				cnxt_otg_defer_work(cnxt_otg, WORK_HOST_RESUME);
			cnxt_bstat = 0;
		}
	} else {
		/* if user unplugged mini-A end of cable,
		 * don't bypass A_WAIT_VFALL.
		 */
		if (cnxt_otg->otg.default_a) {
			switch (state) {
			default:
				cnxt_otg->otg.state = OTG_STATE_A_WAIT_VFALL;
				break;
			case OTG_STATE_A_WAIT_VFALL:
				state = OTG_STATE_A_IDLE;
				/* khubd may take a while to notice and
				 * handle this disconnect, so don't go
				 * to B_IDLE quite yet.
				 */
				break;
			case OTG_STATE_A_IDLE:
				host_suspend(cnxt_otg);
				break;
			case OTG_STATE_B_IDLE:
				break;
			}
		}
		

		
	}

	if (state != cnxt_otg->otg.state)
		pr_debug("  cnxt_otg, %s -> %s\n",
				state_string(state), state_name(cnxt_otg));


}

/*-------------------------------------------------------------------------*/

void  cnxt_otg_clear_latch(struct cnxt_otg_trans *cnxt_otg)
{
	
	u32 cmd;

	cmd = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));

	cmd |= OTG_DP_IS;

	(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;  /* Write to clear the condition */

}

static void
cnxt_otg_work(void *data)
{
	struct cnxt_otg_trans *cnxt_otg = data;
	int		stop;

	/* implicit lock:  we're the only task using this device */
	cnxt_otg->working = 1;
	do {
		stop = test_bit(WORK_STOP, &cnxt_otg->todo);

	
		if (test_and_clear_bit(WORK_UPDATE_OTG, &cnxt_otg->todo)) {
			cnxt_otg_clear_latch(cnxt_otg);
			cnxt_update_otg(cnxt_otg);
			
		}

		if (test_and_clear_bit(WORK_HOST_RESUME, &cnxt_otg->todo)) {
			u32	otg_ctrl;

			/*
			 * skip A_WAIT_VRISE; hc transitions invisibly
			 * skip A_WAIT_BCON; same.
			 */
			switch (cnxt_otg->otg.state) {
			case OTG_STATE_A_WAIT_BCON:
			case OTG_STATE_A_WAIT_VRISE:
				cnxt_otg->otg.state = OTG_STATE_A_HOST;
				pr_debug("  --> a_host\n");
				
				break;
			case OTG_STATE_B_WAIT_ACON:
				cnxt_otg->otg.state = OTG_STATE_B_HOST;
				pr_debug("  --> b_host (acon)\n");
				break;
			case OTG_STATE_B_HOST:
			case OTG_STATE_B_IDLE:
			case OTG_STATE_A_IDLE:
				break;
			default:
				pr_debug("  host resume in %s\n",
						state_name(cnxt_otg));
			}
			host_resume(cnxt_otg);
			// mdelay(10);
			
		}

		
		if (cnxt_otg->todo)
			dev_vdbg("work done, todo = 0x%lx\n",
				cnxt_otg->todo);
		if (stop) {
			printk("stop\n");
			break;
		}
	} while (cnxt_otg->todo);
	cnxt_otg->working = 0;
}


/*-------------------------------------------------------------------------*/




/*-------------------------------------------------------------------------*/



#ifdef	CONFIG_USB_OTG

static int cnxt_otg_enable(struct cnxt_otg_trans *cnxt_otg)
{


	u32 cmd;


	otg_init(cnxt_otg);


	cmd   = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));

	cmd = OTG_DPIE_IE | OTG_BSE_IE | OTG_BSV_IE | OTG_ASV_IE | OTG_AVV_IE | OTG_ID_IE;

	(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;

	
	dev_info("ready for dual-role USB ...\n");

	return 0;
}

#endif

/* add or disable the host device+driver */
static int
cnxt_otg_set_host(struct otg_transceiver *otg, struct usb_bus *host)
{
	struct cnxt_otg_trans	*cnxt_otg = container_of(otg, struct cnxt_otg_trans, otg);

	if (!otg )
		return -ENODEV;

	if (!host) {
		cnxt_otg->otg.host = 0;
		power_down(cnxt_otg);
		return 0;
	}

#ifdef	CONFIG_USB_OTG
	cnxt_otg->otg.host = host;
	dev_dbg("registered host\n");
	host_suspend(cnxt_otg); 
	if (cnxt_otg->otg.gadget)
		return cnxt_otg_enable(cnxt_otg);
	return 0;

#elif	!defined(CONFIG_USB_GADGET_CNXT)
	// FIXME update its refcount
	cnxt_otg->otg.host = host;

	return 0;

#else
	dev_dbg("host sessions not allowed\n");
	return -EINVAL;
#endif

}



static int
cnxt_otg_set_peripheral(struct otg_transceiver *otg, struct usb_gadget *gadget)
{
	struct cnxt_otg_trans	*cnxt_otg = container_of(otg, struct cnxt_otg_trans, otg);
	
	if (!otg)
		return -ENODEV;

	if (!gadget) {
		
		usb_gadget_vbus_disconnect(cnxt_otg->otg.gadget);
		cnxt_otg->otg.gadget = 0;
		power_down(cnxt_otg);
		return 0;
	}

#ifdef	CONFIG_USB_OTG
	cnxt_otg->otg.gadget = gadget;
	dev_dbg("registered gadget\n");
	/* gadget driver may be suspended until vbus_connect () */
	if (cnxt_otg->otg.host)
		return cnxt_otg_enable(cnxt_otg);
	return 0;

#elif	!defined(CONFIG_USB_EHCI_HCD) && !defined(CONFIG_USB_EHCI_HCD_MODULE)
	cnxt_otg->otg.gadget = gadget;

#else
	printk("peripheral sessions not allowed\n");
	return -EINVAL;
#endif



	return 0;
}


/*-------------------------------------------------------------------------*/



static int
cnxt_otg_set_power(struct otg_transceiver *dev, unsigned mA)
{
	struct cnxt_otg_trans	*cnxt_otg = container_of(dev, struct cnxt_otg_trans, otg);

	dcd_power(cnxt_otg->otg_controller_id,mA);
  
}


static int
cnxt_otg_start_srp(struct otg_transceiver *dev)
{
	u32 cmd;
	struct cnxt_otg_trans	*cnxt_otg = container_of(dev, struct cnxt_otg_trans, otg);

	if (!dev || cnxt_otg->otg.state != OTG_STATE_B_IDLE)
		return -ENODEV;

	if( (cnxt_otg->state_internal != OTG_STATE_INTERNAL_B_SESSION_END) || (cnxt_otg->state_internal !=OTG_STATE_INTERNAL_UNDEFINED))
		return -ENODEV;
	
	/* Initiate Data Pulse  srp  */

	cmd   = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));

	cmd =  OTG_HADP | OTG_DP;

	(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;


	/* Initiate VBUS SRP */
	
	cmd   = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));

	cmd  |=  OTG_VC;

	(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;


	cnxt_otg->otg.state = OTG_STATE_B_SRP_INIT;
	return 0;
}



static int
cnxt_otg_start_hnp(struct otg_transceiver *dev)
{
#ifdef	CONFIG_USB_OTG
	struct cnxt_otg_trans	*cnxt_otg = container_of(dev, struct cnxt_otg_trans, otg);

	if (!dev )
		return -ENODEV;
	if (cnxt_otg->otg.default_a && (cnxt_otg->otg.host == NULL
			|| !cnxt_otg->otg.host->b_hnp_enable))
		return -ENOTCONN;
	if (!cnxt_otg->otg.default_a && (cnxt_otg->otg.gadget == NULL
			|| !cnxt_otg->otg.gadget->b_hnp_enable))
		return -ENOTCONN;

	/* We want hardware to manage most HNP protocol timings.
	 * So do this part as early as possible...
	 */
	switch (cnxt_otg->otg.state) {
	case OTG_STATE_B_HOST:
		cnxt_otg->otg.state = OTG_STATE_B_PERIPHERAL;
		/* caller will suspend next */
		break;
	case OTG_STATE_A_HOST:

		/* caller must suspend then clear A_BUSREQ */
		usb_gadget_vbus_connect(cnxt_otg->otg.gadget);

	cmd   = (*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id)));

	cmd |=  OTG_HABA;

	(*((volatile unsigned int *)OTG_REG(cnxt_otg->otg_controller_id))) = cmd;

	

		break;
	case OTG_STATE_A_PERIPHERAL:
		/* initiated by B-Host suspend */
		break;
	default:
		return -EILSEQ;
	}
	
	check_state(cnxt_otg, __FUNCTION__);
	return 0;
#else
	/* srp-only */
	return -EINVAL;
#endif
}


// Global Variables


/*-------------------------------------------------------------------------*/

static int cnxt_otg_probe(struct cnxt_otg_trans *cnxt_otg)
{
	int			status;
	
	

#ifdef	CONFIG_USB_OTG
	status = otg_bind(cnxt_otg);
	if (status < 0) {
		printk("can't bind OTG\n");
		goto fail2;
	}
#endif


	cnxt_otg->otg.set_host = cnxt_otg_set_host,
	cnxt_otg->otg.set_peripheral = cnxt_otg_set_peripheral,
	cnxt_otg->otg.set_power = cnxt_otg_set_power,
	cnxt_otg->otg.start_srp = cnxt_otg_start_srp,
	cnxt_otg->otg.start_hnp = cnxt_otg_start_hnp,

	
	status = otg_set_transceiver(&cnxt_otg->otg);
	if (status < 0)
		printk("can't register transceiver, %d\n",
			status);

	cnxt_otg->state_internal = OTG_STATE_INTERNAL_UNDEFINED;

	return 0;
}


/*-------------------------------------------------------------------------*/

static int __init cnxt_otg_init(void)
{

	int status;


	struct cnxt_otg_trans		*cnxt_otg0,*cnxt_otg1;
	
	cnxt_otg0 = kzalloc(sizeof *cnxt_otg0, GFP_KERNEL);
	if (!cnxt_otg0)
		return 0;

	cnxt_otg1 = kzalloc(sizeof *cnxt_otg1, GFP_KERNEL);
	if (!cnxt_otg1)
		return 0;

	// Set the Resource Information for the OTG Controllers

	cnxt_otg0->otg_controller_id = OTG_ID_0;
	cnxt_otg1->otg_controller_id = OTG_ID_1;
	cnxt_otg0->irq = IRQ_USB_OTG0;
	cnxt_otg1->irq = IRQ_USB_OTG1;
	INIT_WORK(&cnxt_otg0->work, cnxt_otg_work, cnxt_otg0); /* Intialize the Workitem with cnxt_otg0 as dummy item */
	
		
	cnxt_otg_probe(cnxt_otg0); //Probe the OTG Controller Zero
	cnxt_otg_probe(cnxt_otg1); //Probe the OTG Controller One


	return 0; 

}
module_init(cnxt_otg_init);

static void __exit cnxt_otg_exit(void)
{
	otg_set_transceiver(0);

}
module_exit(cnxt_otg_exit);

